SystemOrganization addCategory: #'TZ-Olson'!

!DateAndTime methodsFor: '*TZ-Olson' stamp: 'dtl 2/25/2019 08:11'!
asInteger
	"Compatibility with PointInTime"
	^self asSeconds! !

!DateAndTime methodsFor: '*TZ-Olson' stamp: 'dtl 2/24/2019 17:05'!
inLocalTimeZone
	"Answer an equivalent instance with the local offset for tzName"

	"(DateAndTime now inTimeZone: 'Australia/Perth' ) inLocalTimeZone"

	"A check to verify that the default time zone database matches OS settings:

		(DateAndTime now inLocalTimeZone offset) = DateAndTime now offset
			==> true "


	^self copy setLocalOffsetForTimeZone: TimeZoneDatabase default defaultTimeZone.
! !

!DateAndTime methodsFor: '*TZ-Olson' stamp: 'dtl 2/24/2019 18:18'!
inTimeZone: timeZone
	"Answer an equivalent instance with the local offset for timeZone, where
	tzName is an entry in the system time zone database. Local offset is a
	function of time in a time zone rule set, and varies with daylight savings
	time rules for that time zone."

	^self copy setLocalOffsetForTimeZone: timeZone.
! !

!DateAndTime methodsFor: '*TZ-Olson' stamp: 'dtl 2/24/2019 17:11'!
inTimeZoneNamed: tzName
	"Answer an equivalent instance with the local offset for tzName, where
	tzName is an entry in the system time zone database. Local offset is a
	function of time in a time zone rule set, and varies with daylight savings
	time rules for that time zone."

	"TimeZoneDatabase default grepFor: 'Perth' "

	"DateAndTime now inTimeZoneNamed: #CET "
	"DateAndTime now inTimeZoneNamed: #PST8PDT "
	"DateAndTime now inTimeZoneNamed: 'Australia/Perth' "

	^self copy setLocalOffsetForTimeZoneNamed: tzName.
! !

!DateAndTime methodsFor: '*TZ-Olson' stamp: 'dtl 2/24/2019 16:14'!
setLocalOffsetForTimeZone: localTimeTransform
	"Answer self with the local offset changed to the offset for a timezone"

	| rule |
	rule := localTimeTransform zoneRuleFor: self asSeconds.
	^self localOffsetSeconds: rule offsetSeconds.
! !

!DateAndTime methodsFor: '*TZ-Olson' stamp: 'dtl 2/24/2019 16:48'!
setLocalOffsetForTimeZoneNamed: tzName
	"Answer self with the local offset changed to the offset for tzName"

	"DateAndTime now setLocalOffsetForTimeZoneNamed: #CET "
	"DateAndTime now setLocalOffsetForTimeZoneNamed: #PST8PDT "
	"DateAndTime now setLocalOffsetForTimeZoneNamed: 'Australia/Perth' "

	(TimeZoneDatabase default timeZoneFor: tzName)
		ifNotNilDo: [:tz |
			^self setLocalOffsetForTimeZone: tz ].
	self notify: 'no time zone entry for ', tzName asString.
! !

Object subclass: #LeapSecondRuleSet
	instanceVariableNames: 'leapSecondTable'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'TZ-Olson'!

!LeapSecondRuleSet commentStamp: '<historical>' prior: 0!
I represent a set of rules for applying leap second offsets to local
time. Indirectly, I keep track of the number of seconds in a day, which
is usually 86400. Every once in a great while, this number is changed
by plus or minus one second in order to compensate for changes in the
actual physical rotational speed of the Earth. The entries in my leap
second table keep track of the times at which these one second
adjustments are made.

The effect of leap seconds is the same for all time zones, and is
independent of the time zone offsets maintained by a TimeZoneRuleSet.

Leap seconds have little practical impact for most applications, and
many computer systems simply ignore them in the interest of simplicity.
To understand the role of leap seconds on these systems, consider two
identical computers side by side in the same time zone, with their
local system clocks synchronized. If one of the computers understands
time zones and the other does not, the two computers will have a
difference of opinion (about 22 seconds worth) about how many
seconds have elapsed since January 1, 1970 UTC. In most other
respects, they will behave identically.
!

!LeapSecondRuleSet methodsFor: 'testing' stamp: 'dtl 12/28/1999 20:01'!
isEmpty

	^ leapSecondTable isNil or: [leapSecondTable isEmpty]! !

!LeapSecondRuleSet methodsFor: 'accessing' stamp: 'dtl 12/28/1999 20:01'!
leapSecondTable

	^ leapSecondTable! !

!LeapSecondRuleSet methodsFor: 'accessing' stamp: 'dtl 12/28/1999 20:01'!
leapSecondTable: anArrayOfTransitionRules

	leapSecondTable := anArrayOfTransitionRules! !

!LeapSecondRuleSet methodsFor: 'queries' stamp: 'dtl 12/28/1999 20:01'!
leapSecondsAt: aPointInTimeOrPosixSeconds
	"Answer the leap seconds for a point in time (or its integer value)."

	| rule |
	rule := self leapSecondsRuleFor: aPointInTimeOrPosixSeconds asInteger.
	rule isNil
		ifTrue: [^ 0]
		ifFalse: [^ rule at: 2]
! !

!LeapSecondRuleSet methodsFor: 'private' stamp: 'dtl 12/31/1999 19:06'!
leapSecondsRuleFor: posixSeconds
	"Answer the leap second data record for a point in time (an Integer). Assuming
	that elements of leapSecondTable are sorted by transitionTime, this is the last
	such element for which the transition time is less than posixSeconds."

	| recordsBeforeThisTime |
	leapSecondTable isNil ifTrue: [^ nil].
	leapSecondTable isEmpty ifTrue: [^ nil].
	recordsBeforeThisTime := leapSecondTable select: [:e | (e at: 1) < posixSeconds].
	recordsBeforeThisTime isEmpty
		ifTrue: [^ nil]
		ifFalse: [^ recordsBeforeThisTime last]! !

Object subclass: #LocalTimeTransform
	instanceVariableNames: 'timeZoneName'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'TZ-Olson'!

!LocalTimeTransform commentStamp: '<historical>' prior: 0!
I represent a coordinate transformation between a PointInTime (a point on
a continuous, monotonically increasing, infinite time line) and a local time
line as might be used by a wall clock or by the Time and Date classes.

My coordinate transformation is not necessarily a continuous function,
because it embodies rules for setting the local clock forward and back in
accordance with daylight savings time conventions, rules for applying
leap seconds, and doubtless many other calendar oddities.

My concrete subclasses represent time zones for various locations in
the world. They know how to implement a time coordinate transformation
by applying certain rules, or by making inquiries to an operating system
which knows about time zone rules.!

!LocalTimeTransform class methodsFor: 'instance creation' stamp: 'dtl 12/28/1999 20:01'!
defaultType

	^ self forTzDatabase! !

!LocalTimeTransform class methodsFor: 'instance creation' stamp: 'dtl 12/30/1999 12:06'!
for: aTimeZoneNameString

	"LocalTimeTransform for: 'America/Detroit'"
	"LocalTimeTransform for: 'right/America/Detroit'"
	"LocalTimeTransform for: 'nowhere/noplace'"

	^ self for: aTimeZoneNameString pathSeparator: nil
! !

!LocalTimeTransform class methodsFor: 'instance creation' stamp: 'dtl 12/30/1999 18:36'!
for: aTimeZoneNameString pathSeparator: aCharacter

	"LocalTimeTransform for: 'America/Detroit' pathSeparator: $/ "
	"LocalTimeTransform for: 'right/America/Detroit' pathSeparator: $/ "
	"LocalTimeTransform for: 'right-America-Detroit' pathSeparator: $- "

	| db |
	db := TimeZoneDatabase systemDatabase.
	db isNil
		ifTrue: [^ nil]
		ifFalse: [^ db timeZoneFor: aTimeZoneNameString pathSeparator: aCharacter]! !

!LocalTimeTransform class methodsFor: 'instance creation' stamp: 'dtl 2/26/2019 22:29'!
forSystem
	"For Squeak, use the DateAndTime>>localTimeZone, which should normally be
	configured as a proxy pointing back to the local time tranform in the time
	zone database. On other platforms, just use the current local transform in
	the database."

	"LocalTimeTransform forSystem"

	| dtClass |
	^ (dtClass := Smalltalk at: #DateAndTime ifAbsent: [ nil ]) isNil
		ifTrue: [ self perform: #here ]
		ifFalse: [ dtClass perform: #localTimeZone ]! !

!LocalTimeTransform class methodsFor: 'instance creation' stamp: 'dtl 12/28/1999 20:01'!
forTzDatabase

	^ TimeZoneRuleSet! !

!LocalTimeTransform class methodsFor: 'instance creation' stamp: 'dtl 1/3/2000 04:48'!
here
	"Answer the current time zone, or nil if the classes have not been initialized for this system.
	If you are using the TimeZoneDatabase, please edit TimeZoneDatabase class>>defaultLocation
	to answer your preferred timezone."

	"LocalTimeTransform here"

	| thisZone |
	thisZone := self defaultType here.
	thisZone isNil ifTrue: [
		self notify: 'default time zone not found, using ', self hereUsingNothing printString.
		thisZone := self hereUsingNothing].
	^ thisZone! !

!LocalTimeTransform methodsFor: 'timezone conversion' stamp: 'dtl 2/26/2019 21:20'!
date: aDate andTime: aTime
	"Answer a DateAndTime for the specified date and time in this time zone.
	If close to a daylight saving transition, iterate to a solution such that
	offset correctly matches the time zone rules."

	"(LocalTimeTransform for: 'America/Detroit')
		date: Date today
		andTime: Time now"

	| dt offset result |
	dt := DateAndTime date: aDate time: aTime.
	result := dt copy.
	[offset := self localOffsetSecondsAt: result.
	result offsetSeconds = offset]
		whileFalse: [result := self similarLocal: result with: offset].
	^ result
! !

!LocalTimeTransform methodsFor: 'timezone conversion' stamp: 'dtl 2/25/2019 20:46'!
date: aDate andTime: aTime inTimeZone: aTimeZone
	"For aDate and aTime here, answer the corresponding date and time in aTimeZone."

	"(LocalTimeTransform for: 'America/Detroit')
		date: Date today
		andTime: Time now
		inTimeZone: (LocalTimeTransform for: 'America/Los_Angeles')"

	| dt offset |
	dt := self date: aDate andTime: aTime.
	offset := aTimeZone localOffsetSecondsAt: dt asSeconds.
	dt localOffsetSeconds: offset.
	^ Array
		with: dt asDate
		with: dt asTime! !

!LocalTimeTransform methodsFor: 'leap seconds' stamp: 'dtl 12/28/1999 20:01'!
leapSecondsAt: aPointInTime
	"Answer the number of leap seconds since January 1, 1970 UTC as of aPointInTime"

	^ self subclassResponsibility 
! !

!LocalTimeTransform methodsFor: 'offsets from UTC' stamp: 'dtl 12/28/1999 20:01'!
localOffsetSecondsAt: aPointInTime
	"Answer the offset from UTC in this time zone as of aPointInTime"

	^ self subclassResponsibility 
! !

!LocalTimeTransform methodsFor: 'timezone conversion' stamp: 'dtl 2/26/2019 21:34'!
similarLocal: dateAndTime with: offset
	"Answer a DateAndTime in this time zone with the same local representation as aDateAndTime"

	| microsecs |
	( DateAndTime respondsTo: #utcMicroseconds:offset: )
		ifTrue: [ "Recent Chronology-Core package with UTC support"
				microsecs := offset * -1000000 + dateAndTime utcMicroseconds.
				^ DateAndTime perform: #utcMicroseconds:offset: with: microsecs with: offset ]
		ifFalse: [ "Older versions. Result loses subsecond precision"
				^ DateAndTime
						year: dateAndTime year
						day: dateAndTime day
						hour: dateAndTime hour
						minute: dateAndTime minute
						second: dateAndTime second
						offset: offset seconds ].
! !

!LocalTimeTransform methodsFor: 'accessing' stamp: 'dtl 12/28/1999 20:01'!
timeZoneName

	^ timeZoneName! !

!LocalTimeTransform methodsFor: 'accessing' stamp: 'dtl 12/28/1999 20:01'!
timeZoneName: aStringOrSymbol

	timeZoneName := aStringOrSymbol! !

LocalTimeTransform subclass: #TimeZoneRuleSet
	instanceVariableNames: 'tzNameAbbreviations transitionTimeTable leapSecondRuleSet posixTzEnvString reversedTransitonTimeTable'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'TZ-Olson'!

!TimeZoneRuleSet commentStamp: '<historical>' prior: 0!
I represent a timezone for a location in the world. I know the rules
for determining offsets from UTC, and for applying leap seconds.

My rules are constructed from a data set in a tzfile(5) timezone
database file. I know how to load these rules from a file, and how
to apply them to calculate UTC offsets, daylight savings time, and
leap seconds.

A tzfile database file is typically generated by the zic(1) compiler
distributed as part of the public domain timezone database in the
~ftp/pub directory of elsie.nci.nih.gov FTP server.
!

TimeZoneRuleSet subclass: #FixedTimeZoneRuleSet
	instanceVariableNames: 'offset'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'TZ-Olson'!

!FixedTimeZoneRuleSet commentStamp: 'PaulDeBruicker 3/14/2016 08:24' prior: 0!
Used for the tz rules in Etc/GMT* 

These timezones have constant offsets with no DST adjustments.!

!FixedTimeZoneRuleSet methodsFor: 'as yet unclassified' stamp: 'PaulDeBruicker 3/13/2016 22:34'!
localOffsetSecondsAt: aPointInTimeOrPosixSeconds
	^ offset! !

!FixedTimeZoneRuleSet methodsFor: 'accessing' stamp: 'PaulDeBruicker 3/13/2016 22:53'!
offset: anInteger 
	offset := anInteger! !

!TimeZoneRuleSet class methodsFor: 'instance creation' stamp: 'dtl 1/3/2000 04:36'!
here
	"Answer an existing instance from the system time zone database"

	"TimeZoneRuleSet here"

	| db |
	db := TimeZoneDatabase systemDatabase.
	db isNil
		ifTrue: [^ nil]
		ifFalse: [^ TimeZoneDatabase systemDatabase defaultTimeZone]! !

!TimeZoneRuleSet class methodsFor: 'instance creation' stamp: 'dtl 12/28/1999 20:01'!
loadFrom: aTzFileName
	"Load an instance of me from a tzfile file."

	^ self loadFrom: aTzFileName prefixPath: nil
! !

!TimeZoneRuleSet class methodsFor: 'instance creation' stamp: 'dtl 1/1/2000 14:31'!
loadFrom: aTzFileName prefixPath: aString
	"Load an instance of me from a tzfile file. Warning: Must use a StandardFileStream,
	because CrLrFileStream gets confused and tries to treat the file as binary."

	"TimeZoneRuleSet loadFrom: 'America/Detroit' prefixPath: '/usr/share/zoneinfo/'"
	"TimeZoneRuleSet loadFrom: 'right/America/Detroit' prefixPath: '/usr/share/zoneinfo/'"
	"TimeZoneRuleSet loadFrom: 'America/Barbados' prefixPath: '/usr/share/zoneinfo/'"
	"TimeZoneRuleSet loadFrom: 'America/Chicago' prefixPath: '/usr/share/zoneinfo/'"
	"TimeZoneRuleSet loadFrom: 'Europe/Zurich' prefixPath: '/usr/share/zoneinfo/'"

	^ (TzFileLoader forFile: aTzFileName prefixPath: aString) load at: 1
! !

!TimeZoneRuleSet methodsFor: 'accessing' stamp: 'dtl 11/14/2004 07:39'!
abbreviation
	"Answer one of possibly several abbreviations for this time zone"

	^ self tzNameAbbreviations anyOne asString! !

!TimeZoneRuleSet methodsFor: 'accessing' stamp: 'dtl 2/26/2019 22:58'!
dstTransitionTimes

	^ self transitionTimeTable
		collect: [:rule | DateAndTime fromUnixTime: rule transitionTime]! !

!TimeZoneRuleSet methodsFor: 'comparing' stamp: 'dtl 12/30/1999 10:02'!
isEquivalentTo: aTimeZoneRuleSet
	"Answer true if aTimeZoneRuleSet has a transition table the same as
	that of the receiver."

	(aTimeZoneRuleSet isKindOf: LocalTimeTransform)
		ifTrue: [^ self transitionTimeTable = aTimeZoneRuleSet transitionTimeTable]
		ifFalse: [^ false]! !

!TimeZoneRuleSet methodsFor: 'accessing' stamp: 'dtl 12/28/1999 20:01'!
leapSecondRuleSet

	^ leapSecondRuleSet! !

!TimeZoneRuleSet methodsFor: 'accessing' stamp: 'dtl 12/28/1999 20:01'!
leapSecondRuleSet: aLeapSecondRuleSet
	"aLeapSecondRuleSet answers leap seconds at a PointInTime. Normally, there is
	one LeapSecondRuleSet which applies to all time zones, and this LeapSecondRuleSet
	is held by the system TimeZoneDatabase."

	leapSecondRuleSet := aLeapSecondRuleSet! !

!TimeZoneRuleSet methodsFor: 'accessing' stamp: 'dtl 2/26/2019 23:04'!
leapSecondTransitionTimes

	self leapSecondRuleSet ifNil: [^ #()].
	^ (self leapSecondRuleSet leapSecondTable
		collect: [:r | DateAndTime fromUnixTime: r first]) asArray! !

!TimeZoneRuleSet methodsFor: 'leap seconds' stamp: 'dtl 12/28/1999 20:01'!
leapSecondsAt: aPointInTimeOrPosixSeconds
	"Answer the leap seconds for a point in time (or its integer value)."

	leapSecondRuleSet isNil
		ifTrue: [^ 0]
		ifFalse: [^ leapSecondRuleSet leapSecondsAt: aPointInTimeOrPosixSeconds]

! !

!TimeZoneRuleSet methodsFor: 'offsets from UTC' stamp: 'dtl 1/8/2000 13:52'!
localOffsetSecondsAt: aPointInTimeOrPosixSeconds
	"Answer the offset seconds for a point in time (or its integer value)."

	"TimeZoneRuleSet here localOffsetSecondsAt: PointInTime now"

	| rule |
	rule := self zoneRuleFor: aPointInTimeOrPosixSeconds asInteger.
	rule isNil
		ifTrue: ["no timezone rules, default to zero offset"
				^ 0]
		ifFalse: [^ rule offsetSeconds]
! !

!TimeZoneRuleSet methodsFor: 'accessing' stamp: 'dtl 11/13/2004 22:59'!
name
	"Name when used as a TimeZone"

	^ (self timeZoneName ifNil: [super name]) asString! !

!TimeZoneRuleSet methodsFor: 'accessing' stamp: 'dtl 5/31/2009 13:19'!
posixTzEnvString
	"Answer a newline-enclosed, POSIX-TZ-environment-variable-style string for use in
	handling instants after the last transition time stored in the file (an empty string
	if there is no POSIX representation for such instants). This variable is set only in
	version 2 or greater tzinfo files, so may be nil for older tzfile inputs, or an empty
	string if using newer format tzinfo when there is no POSIX representation for such
	instants for this rule set."

	^ posixTzEnvString! !

!TimeZoneRuleSet methodsFor: 'accessing' stamp: 'dtl 5/31/2009 13:15'!
posixTzEnvString: aByteString 

	posixTzEnvString := aByteString! !

!TimeZoneRuleSet methodsFor: 'printing' stamp: 'dtl 12/28/1999 20:01'!
printOn: aStream

	super printOn: aStream.
	aStream nextPutAll: ' for '.
	aStream nextPutAll: self timeZoneName! !

!TimeZoneRuleSet methodsFor: 'accessing' stamp: 'PaulDeBruicker 3/13/2016 12:31'!
reversedTransitionTimeTable
	^ reversedTransitonTimeTable ifNil: [ reversedTransitonTimeTable := transitionTimeTable reverse ]! !

!TimeZoneRuleSet methodsFor: 'accessing' stamp: 'dtl 12/28/1999 20:01'!
transitionTimeTable

	^ transitionTimeTable! !

!TimeZoneRuleSet methodsFor: 'accessing' stamp: 'PaulDeBruicker 3/13/2016 12:46'!
transitionTimeTable: anArrayOfArrays
	transitionTimeTable := anArrayOfArrays.
	reversedTransitonTimeTable := nil! !

!TimeZoneRuleSet methodsFor: 'accessing' stamp: 'dtl 12/28/1999 20:01'!
tzNameAbbreviations

	^ tzNameAbbreviations! !

!TimeZoneRuleSet methodsFor: 'accessing' stamp: 'dtl 12/28/1999 20:01'!
tzNameAbbreviations: anArrayOfStrings

	tzNameAbbreviations := anArrayOfStrings! !

!TimeZoneRuleSet methodsFor: 'private' stamp: 'PaulDeBruicker 3/13/2016 16:09'!
zoneRuleFor: posixSeconds
	"Answer the zone data record for a point in time (an Integer). Assuming
	that elements of transitionTimeTable are sorted by transitionTime, this is the last
	such element for which transitionTime is less than posixSeconds."

	transitionTimeTable isEmptyOrNil
		ifTrue: [ ^ nil ].
	self reversedTransitionTimeTable
		do: [ :e | 
			e transitionTime <= posixSeconds
				ifTrue: [ ^ e ] ].
	^ transitionTimeTable first	"Default to first rule in the list"! !

Object subclass: #TimeZoneDatabase
	instanceVariableNames: 'defaultLocation defaultTimeZone timeZones leapSecondRuleSet indexSeparator'
	classVariableNames: 'OsPathNameSeparator Platform ThisSystemDatabase'
	poolDictionaries: ''
	category: 'TZ-Olson'!

!TimeZoneDatabase commentStamp: '<historical>' prior: 0!
I represent a collection of time zone rules for many time zones. I know how
to determine the number of seconds offset from UTC for any of my time zones
at any time within the range of my time zone rules, as well as the number of
leap seconds since the origin of time for any point in time. I know how to build
time zone rules by reading compiled tzfile files from an external source.

I store time zone information in a hierarchy of Dictionaries, where each Dictionary
contains a group of related TimeZoneRuleSets and possibly other Dictionaries. This
organization is similar to that of a tree structured directory of files and is intended
to reflect the organization of tzfile time zone files as found on many computer systems.
TimeZoneRuleSets are indexed by a path name consisting of a string with separator
characters, for example 'right/America/Detroit' or 'right-America-Detroit'.

I store leap second information in a single LeapSecondRuleSet which may optionally
be referenced by TimeZoneRuleSets which choose to honor leap second conventions.
On many systems, time zones which honor leap second conventions are given names
which begin with 'right', such as 'right/America/Detroit'.

I keep track of two types of rules. First, time zone offset rules are used to determine
the difference in seconds between UTC time, and time as represented in a specified
time zone. For example, there might be a five hour difference between time
as displayed locally and UTC time; this difference is represented by an offset
of -18000 seconds from UTC. These time zone offset rules also embody daylight
savings time rules, which can cause the offset to change twice annually in
accordance with daylight savings time clock setting conventions.

Second, leap second rules are used to apply minor offsets to both UTC time and
local time. The small offsets are specified such that time as displayed on a wall
clock time (in any time zone) stays synchronized with the physical rotation of
the Earth.

Leap second offsets are frequently ignored in the interest of simplifying the
mapping of absolute time (represented as class PointInTime) to the local
representation of time (which might appear as a wall clock display, or as the
date and time displayed in Smalltalk as Time>>dateAndTimeNow). In any case,
the leap second rules are common to all time zones, even though they are
redundantly encoded in any tzfile formatted time zone file (this is just an
implementation convenience for the time zone software which uses tzfile
time zone data).

Depending on the way in which a tzfile time zone file was compiled (using
the zic(1) program), it may or may not have leap second rules included.
Many computer systems will use time zone files without leap seconds in the
interest of simplicity, even though the time mapping is not strictly correct.

!

!TimeZoneDatabase class methodsFor: 'setting local time zone' stamp: 'dtl 2/24/2019 18:05'!
allTimeZoneNamesThatMatchMyOwn
	"Names for all time zones for which local offset matches my local offset now"

	"TimeZoneDatabase allTimeZoneNamesThatMatchMyOwn"

	^self allTimeZonesThatMatchMyOwn collect: [ :e | e name ]
! !

!TimeZoneDatabase class methodsFor: 'setting local time zone' stamp: 'dtl 2/24/2019 18:05'!
allTimeZonesThatMatchMyOwn
	"All time zones for which local offset matches my local offset now"

	"TimeZoneDatabase allTimeZonesThatMatchMyOwn"

	| now currentOffset seconds |
	now := DateAndTime now.
	currentOffset := now offset asSeconds.
	seconds := now asSeconds.
	^self default allTimeZones
		select: [ :zone |
			(zone localOffsetSecondsAt: seconds) = currentOffset ].
! !

!TimeZoneDatabase class methodsFor: 'instance creation' stamp: 'dtl 2/15/2007 21:31'!
buildDefaultUTCDatabase
	"Build a minimal timezone database and install it as ThisSystemDatabase."

	"TimeZoneDatabase buildDefaultUTCDatabase"

	| db |
	db := TimeZoneDatabase basicNew instVarAt: 1 put: 'UTC';
			instVarAt: 2 put: nil;
			instVarAt: 3 put: (Dictionary new add: #UTC -> (TimeZoneRuleSet basicNew instVarAt: 1 put: #UTC;
						instVarAt: 2 put: #(#UTC );
						instVarAt: 3 put: #();
						instVarAt: 4 put: nil;
						yourself);
				yourself);
			instVarAt: 4 put: nil;
			instVarAt: 5 put: $/;
			yourself.
	self thisSystemDatabase: db.
	^ ThisSystemDatabase! !

!TimeZoneDatabase class methodsFor: 'instance creation' stamp: 'dtl 11/13/2004 13:31'!
buildSystemDatabase
	"Build a full timezone database and install it as ThisSystemDatabase. Load the
	database from any tzfile files found in the default location for this system."

	"TimeZoneDatabase buildSystemDatabase"

	self thisSystemDatabase: (self fromZoneinfoDirectory: self tzPrefixPath).
	^ ThisSystemDatabase

! !

!TimeZoneDatabase class methodsFor: 'instance creation' stamp: 'PaulDeBruicker 2/13/2016 16:46'!
buildSystemDatabaseForSqueak
	"Build a full timezone database and install it as ThisSystemDatabase. Load the
	database from any tzfile files found in the default location for this system.
	Check first to determine if the time zone files exist, and issue a warning if
	the files are not available."

	"TimeZoneDatabase buildSystemDatabaseForSqueak"
|entries|
entries :=self platformIsSqueak ifTrue:[((Smalltalk at:#FileDirectory )on: self tzPrefixPath) entries] ifFalse:[self tzPrefixPath asFileReference entries].
	^  entries isEmpty
		ifTrue: [ self notify: 'no time zone files found in ' , self tzPrefixPath ]
		ifFalse: [ self buildSystemDatabase ]! !

!TimeZoneDatabase class methodsFor: 'instance creation' stamp: 'dtl 11/13/2004 13:32'!
buildSystemDatabaseFromExamples
	"Build a full timezone database and install it as ThisSystemDatabase"

	"TimeZoneDatabase buildSystemDatabaseFromExamples"

	self thisSystemDatabase: (self fromExampleFiles).
	^ ThisSystemDatabase

! !

!TimeZoneDatabase class methodsFor: 'private' stamp: 'PaulDeBruicker 10/7/2011 10:36'!
changeSeparatorFrom: aCharacter to: anotherCharacter in: aString
	"Replace the path separator aCharacter with anotherCharacter in aString."

	(self platformIsSqueakOrPharo ) ifTrue:
		[^ aString copyReplaceAll: (String with: aCharacter) with: (String with: anotherCharacter)].
	(self platform == #VisualWorks) ifTrue: 
		[^ aString copyReplaceAll: (String with: aCharacter) with: (String with: anotherCharacter)].
 	(self platform == #'Smalltalk/X') ifTrue: 
		[^ aString copy replaceAll: aCharacter with: anotherCharacter].
! !

!TimeZoneDatabase class methodsFor: 'private' stamp: 'PaulDeBruicker 10/7/2011 10:37'!
dateFromString: aString
	"Answer an instance of created from a string with format DD.MM.YYYY."

	(self platformIsSqueakOrPharo ) ifTrue: [^ Date fromString: aString].
	(self platform == #VisualWorks) ifTrue: [^ Date readFrom: (ReadStream on: aString)].
	(self platform == #'Smalltalk/X') ifTrue: [^ Date readFrom: (ReadStream on: aString)].
! !

!TimeZoneDatabase class methodsFor: 'defaulting' stamp: 'dtl 2/24/2019 20:19'!
default
	^self systemDatabase! !

!TimeZoneDatabase class methodsFor: 'defaulting' stamp: 'dtl 8/2/2008 12:17'!
defaultLocation
	"Default name string to use as the key for the system timezone database. The system
	default timezone can be set explicitly with TimeZoneDatabase>>defaultLocation:
	This class side method is used only if the database is not explicitly initialized to the
	correct local time zone name. Edit this to suit your system."

	^ 'UTC'		"Detroit time zone with leap second table"
! !

!TimeZoneDatabase class methodsFor: 'initialize-release' stamp: 'dtl 2/24/2019 22:23'!
defaultLocation: tzName

	"Query for suitable TZ names:
		TimeZoneDatabase default grepFor: 'Detroit' "

	"Set default to a chosen value:
		TimeZoneDatabase defaultLocation: 'America/Detroit' "

	self systemDatabase defaultLocation: tzName.! !

!TimeZoneDatabase class methodsFor: 'defaulting' stamp: 'dtl 2/25/2019 08:07'!
defaultTimeZone
	^self systemDatabase defaultTimeZone! !

!TimeZoneDatabase class methodsFor: 'private' stamp: 'PaulDeBruicker 2/13/2016 16:28'!
directoryContents: aDirectoryPathName
	"Answer the contents of a directory named aDirectoryPathName, in the
	form of a two element array. The first element of the result array is
	a collection of all non-directory elements in aDirectoryPathName (usually
	regular files), and the second element of the result array is a collection 
	of names of directories contained in aDirectoryPathName. Implementation
	is Smalltalk platform specific, so the method is compiled into a block which
	can be evaluated with aDirectoryPathName as its argument."

	"TimeZoneDatabase directoryContents: '/usr/share/zoneinfo' "

	| path result |
	result := Array with: OrderedCollection new with: OrderedCollection new.
	aDirectoryPathName last == self separator
		ifTrue: [ path := aDirectoryPathName ]
		ifFalse: [ path := aDirectoryPathName , (String with: self separator) ].
	self platformIsSqueakOrPharo
		ifTrue: [ 
			| entries |
			entries := self platformIsSqueak
				ifTrue: [ ((Smalltalk at: #FileDirectory) forFileName: path) entries ]
				ifFalse: [ path asFileReference entries ].
			entries
				do: [ :e | 
					e isDirectory
						ifTrue: [ (result at: 2) add: e name ]
						ifFalse: [ (result at: 1) add: e name ] ] ]
		ifFalse: [ 
			((Smalltalk at: #Filename) named: path) directoryContents
				do: [ :e | 
					((Smalltalk at: #Filename) named: path , (String with: TimeZoneDatabase separator) , e) isDirectory
						ifTrue: [ (result at: 2) add: e ]
						ifFalse: [ (result at: 1) add: e ] ] ].
	^ result! !

!TimeZoneDatabase class methodsFor: 'private' stamp: 'dtl 1/2/2000 18:20'!
directoryFilesIn: aPathName inBasePath: prefixPath
	"Search a directory recursively and answer a collection of all directory files."

	"TimeZoneDatabase directoryFilesIn: '' inBasePath: '/usr/share/zoneinfo/'"

	| initialPath contents result relativeName |
	initialPath := self pathFromPrefix: prefixPath andName: aPathName.
	contents := self directoryContents: initialPath.
	result := OrderedCollection new: (contents at: 1) size.
	(contents at: 2) do: [:e |
		relativeName := (self pathFromPrefix: aPathName andName: e).
		result add: relativeName.
		result addAll: (self directoryFilesIn: relativeName inBasePath: prefixPath)].
	^ result

! !

!TimeZoneDatabase class methodsFor: 'database verification' stamp: 'dtl 1/3/2000 21:37'!
dumpRulesToFilesIn: aDirectoryName
	"Dump timezone rules into external files in a format which allows them
	to be compared to the output of the zdump(1) reference program. The
	file 'America.Detroit' (for example) in the dump directory can then be
	compared with the output of the '/usr/sbin/zdump -v America/Detroit'
	command."

	"Time millisecondsToRun: [TimeZoneDatabase dumpRulesToFilesIn: '/tmp/zonedump/']"

	| f n zones idx |
	idx := 0.
	zones := self systemDatabase allTimeZones.
	('dumping zone data to ', aDirectoryName)
		displayProgressAt: Sensor cursorPoint
		from: 1 to: zones size
		during: [ :bar |
			zones do: [:ruleSet |
			idx := idx + 1.
			bar value: idx.
			n := self changeSeparatorFrom: $/ to: $. in: ruleSet name.
			f := FileStream fileNamed: (aDirectoryName, n).
			ruleSet transitionTimeTable do: [:e |
				e printVerboseOn: f withLeapSeconds: (ruleSet leapSecondsAt: e transitionTime).
				f nextPut: Character lf].
				f close]]

! !

!TimeZoneDatabase class methodsFor: 'defaulting' stamp: 'dtl 1/8/2000 18:36'!
exampleTzFiles
	"Answer an array of names of some compiled timezone files, provided as examples.
	The names here use a naming convention similar to that used on systems which
	support tzfile timezone databases, except that the path separator is changed such
	that 'America/Detroit' becomes 'America-Detroit'.

	By convention, a timezone file name which begins with 'right', such as
	'right-America-Detroit', is the version which contains leap second rules. The
	'posix-America-Detroit' and 'America-Detroit' versions are identical, and contain
	no leap second rules."

	^ #('America-Buenos_Aires'
		'America-Detroit'
		'America-Los_Angeles'
		'Europe-Berlin'
		'Europe-London'
		'Europe-Paris'
		'posix-America-Buenos_Aires'
		'posix-America-Detroit'
		'posix-America-Los_Angeles'
		'posix-Europe-Berlin'
		'posix-Europe-London'
		'posix-Europe-Paris'
		'right-America-Buenos_Aires'
		'right-America-Detroit'
		'right-America-Los_Angeles'
		'right-Europe-Berlin'
		'right-Europe-London'
		'right-Europe-Paris')

! !

!TimeZoneDatabase class methodsFor: 'instance creation' stamp: 'dtl 12/30/1999 20:11'!
fromExampleFiles
	"Build a full timezone database from a small set of example tzfile files. Attempt
	to save space to avoiding duplicate entries for leap second rules and time zone
	offset rule sets."

	"TimeZoneDatabase fromExampleFiles"

	^ self fromFiles: self exampleTzFiles prefixPath: self tzExamplesPrefixPath pathSeparator: $-

! !

!TimeZoneDatabase class methodsFor: 'instance creation' stamp: 'dtl 12/28/1999 20:01'!
fromFiles: aCollectionOfNames prefixPath: aPathName

	"TimeZoneDatabase fromFiles: TimeZoneDatabase exampleTzFiles prefixPath: TimeZoneDatabase tzExamplesPrefixPath"

	^ self fromFiles: aCollectionOfNames prefixPath: aPathName pathSeparator: nil

! !

!TimeZoneDatabase class methodsFor: 'instance creation' stamp: 'dtl 12/31/1999 18:47'!
fromFiles: aCollectionOfNames prefixPath: aPathName pathSeparator: aCharacter
	"Build a full timezone database from a set of tzfile files. The path names have
	aPathName prepended in order to permit the names (without aPathName) to be
	used as the names of the TimeZoneRuleSets. Attempt to save space by avoiding
	duplicate entries for leap second rules and time zone offset rule sets.

	The path separator (aCharacter) is always replaced by self>>indexSeparator for
	purposes of naming the time zones in the database. This enforces a consistent
	time zone path name independent of the system platform."

	"TimeZoneDatabase fromFiles: TimeZoneDatabase exampleTzFiles prefixPath: TimeZoneDatabase tzExamplesPrefixPath pathSeparator: $- "

	| db idx |
	db := super new.
	idx := 0.
	self showProgressFrom: 1 to: aCollectionOfNames size withMessage: 'loading timezone data files'
		forBlock: [ :bar |
			aCollectionOfNames do: [:e |
				idx := idx + 1.
				bar value: idx.
				db loadDataFrom: e
					prefixPath: aPathName
					pathSeparator: aCharacter
					condenseDuplicates: false]].
	^ db normalize
! !

!TimeZoneDatabase class methodsFor: 'instance creation' stamp: 'dtl 1/2/2000 17:01'!
fromZoneinfoDirectory: aPathName
	"Build a full timezone database for a compiled set of timezone files as found
	in the directory named aPathName."

	"Time millisecondsToRun: [(TimeZoneDatabase fromZoneinfoDirectory: '/usr/share/zoneinfo') inspect]"

	| path names |
	path := aPathName, (String with: self separator).
	names := self tzFilesInDirectoryTree: path.
	^ self fromFiles: names prefixPath: path

! !

!TimeZoneDatabase class methodsFor: 'setting local time zone' stamp: 'dtl 11/9/2004 19:23'!
grepFor: aString
	"Convenience method to help find time zones for locations, especially if
	the name of the key city is known."

	"TimeZoneDatabase grepFor: 'Moscow' "
	"TimeZoneDatabase grepFor: 'Detroit' "
	"TimeZoneDatabase grepFor: 'UTC' "
	"TimeZoneDatabase grepFor: 'GMT' "

	^ self systemDatabase grepFor: aString
! !

!TimeZoneDatabase class methodsFor: 'defaulting' stamp: 'dtl 12/30/1999 18:40'!
indexSeparator
	"Answer the character which is used as a path separator for the time zone names.
	The default is to a Unix style path separator, which happens to coincide with the
	naming convention for tzfile time zone files distributed with may Unix type systems."

	^ $/! !

!TimeZoneDatabase class methodsFor: 'initialize-release' stamp: 'dtl 2/24/2019 18:47'!
initialize
	"Attempt to load some sample database files in order to provide a system database.
	Different flavors of Smalltalk require different initialization routines, so the
	required code is compiled for each supported platform."

	"TimeZoneDatabase initialize"

	| contents |
	Platform := self platformSymbol.
	self platformIsSqueakOrPharo  ifTrue: [self initializeForSqueak].
	Platform == #VisualWorks ifTrue: [
		contents := ((Smalltalk at: #Filename) currentDirectory directoryContents).
		(TimeZoneDatabase exampleTzFiles select: [:e | (contents includes: e) not]) isEmpty
			ifTrue: [TimeZoneDatabase buildSystemDatabaseFromExamples]
			ifFalse: [TimeZoneDatabase buildDefaultUTCDatabase]].
	Platform == #'Smalltalk/X' ifTrue: [
		contents := ((Smalltalk at: #Filename) currentDirectory directoryContents).
		(TimeZoneDatabase exampleTzFiles select: [:e | (contents includes: e) not]) isEmpty
			ifTrue: [TimeZoneDatabase buildSystemDatabaseFromExamples]
			ifFalse: [TimeZoneDatabase buildDefaultUTCDatabase]]
! !

!TimeZoneDatabase class methodsFor: 'initialize-release' stamp: 'dtl 3/2/2019 20:13'!
initializeForSqueak
	"If time zone files are installed on this system, load them into a system
	database, otherwise use a minimal UTC database."

	| fd |
	[ 
	fd := self platformIsSqueak
		ifTrue: [ (Smalltalk at: #FileDirectory) on: TimeZoneDatabase tzPrefixPath ]
		ifFalse: [ TimeZoneDatabase tzPrefixPath asFileReference ] ]
		on: Error
		do: [ 
			"no such directory"
			 ].
	(fd isNil or: [ fd entries isEmpty ])
		ifTrue: [ TimeZoneDatabase buildDefaultUTCDatabase ]
		ifFalse: [ TimeZoneDatabase buildSystemDatabaseForSqueak ].
! !

!TimeZoneDatabase class methodsFor: 'private' stamp: 'dtl 1/2/2000 22:23'!
isTzFileType: aPathName

	"TimeZoneDatabase isTzFileType: '/usr/share/zoneinfo/America/Detroit' "

	| stream |
	stream := (TzFileLoader readStreamFor: aPathName) binary.
	((stream next: 4) = 'TZif' asByteArray)
		ifTrue: [stream close. ^ true]
		ifFalse: [stream close. ^ false]
! !

!TimeZoneDatabase class methodsFor: 'examples' stamp: 'dtl 2/24/2019 17:56'!
partyTimes
	"Answer the times, in my own time zone, at which the New Year's Eve celebration
	for Year 2000 will occur for all time zones in the database. Sort the results so as to
	facilitate recognition of the grand events by committed revelers in my own time zone.

	Editor's note: The original TimeZoneDatabase package was written in 1999, inspired
	by the hysteria of the Y2K crisis during which the End Of Days was anticipated by
	programmers, business managers, and lay persons who believed themselves hopelessly
	trapped by the legacy of lesser software systems. It seemed natural at the time that
	Squeakers should handle the matter in a more upbeat manner."

	"TimeZoneDatabase partyTimes"

	^ self partyTimesForNewYear: 2000
! !

!TimeZoneDatabase class methodsFor: 'examples' stamp: 'dtl 2/26/2019 21:52'!
partyTimesForNewYear: newYear
	"Answer the times, in my own time zone, at which the New Year's Eve celebration
	for the newYear will occur for all time zones in the database. Sort the results so as to
	facilitate recognition of the grand events by committed revelers in my own time zone."

	"TimeZoneDatabase partyTimesForNewYear: 2005"

	| db myTimeZone c dateAndTime seconds result |
	db := TimeZoneDatabase systemDatabase.
	db isNil ifTrue: [
		self notify: 'no system time zone database'.
		^ nil].
	myTimeZone := LocalTimeTransform here.
	c := SortedCollection sortBlock: [:x :y | (x at: 3) <= (y at: 3)].
	db allTimeZones do: [:e |
		dateAndTime := e
			date: (self dateFromString: '01.01.', newYear asString)
			andTime: (Time fromSeconds: 0)
			inTimeZone: myTimeZone.
		seconds := (dateAndTime at: 1) asSeconds + (dateAndTime at: 2) asSeconds.
		c add: { e . dateAndTime . seconds } ].
	result := OrderedCollection new.
	c do: [:e | result add: (Array with: (e at: 1) timeZoneName with: (e at: 2))].
	^ result asArray

! !

!TimeZoneDatabase class methodsFor: 'private' stamp: 'dtl 1/2/2000 13:03'!
pathFromPrefix: prefixString andName: aString
	"Construct a path name, adding path separator if needed."

	^ (prefixString isNil or: [prefixString size == 0])
		ifTrue: [aString]
		ifFalse: [(prefixString last == self separator)
			ifTrue: [prefixString, aString]
			ifFalse: [prefixString, (String with: self separator), aString]]
! !

!TimeZoneDatabase class methodsFor: 'private' stamp: 'dtl 2/24/2019 18:48'!
platform
	"Answer what version of Smalltalk we are using."

	"TimeZoneDatabase platform"

	^Platform
! !

!TimeZoneDatabase class methodsFor: 'testing' stamp: 'PaulDeBruicker 2/13/2016 16:35'!
platformIsSqueak
	^ self platform == #Squeak ! !

!TimeZoneDatabase class methodsFor: 'testing' stamp: 'PaulDeBruicker 10/7/2011 10:34'!
platformIsSqueakOrPharo
	^self platform == #Squeak or: [self platform == #Pharo]! !

!TimeZoneDatabase class methodsFor: 'private' stamp: 'dtl 2/24/2019 18:44'!
platformSymbol
	"Answer what version of Smalltalk we are using."

	| versionString versionClass |
	(versionClass := Smalltalk at: #SystemVersion ifAbsent: []) isNil
		ifTrue: [versionString := Smalltalk version]
		ifFalse: [versionString := versionClass current version].
	('Squeak*' match: versionString) ifTrue: [^ #Squeak].
	('Pharo*' match: versionString) ifTrue: [^ #Pharo].
	('VisualWorks*' match: versionString) ifTrue: [^ #VisualWorks].
	('*Smalltalk.st*' match: versionString) ifTrue: [^ #'Smalltalk/X'].
	^ #unknown! !

!TimeZoneDatabase class methodsFor: 'private' stamp: 'PaulDeBruicker 2/13/2016 16:38'!
separator
	"Answer the platform-specific file path name separator"

	"TimeZoneDatabase separator"

	OsPathNameSeparator isNil
		ifTrue: [ 
			self platformIsSqueakOrPharo
				ifTrue: [ 
					self platformIsSqueak
						ifTrue: [ OsPathNameSeparator := (Smalltalk at: #FileDirectory) pathNameDelimiter ]
						ifFalse: [ OsPathNameSeparator := (Smalltalk at: #FileSystem) disk delimiter ] ]
				ifFalse: [ OsPathNameSeparator := (Smalltalk at: #Filename) separator ] ].
	^ OsPathNameSeparator! !

!TimeZoneDatabase class methodsFor: 'setting local time zone' stamp: 'dtl 2/25/2019 08:18'!
setLocalTimeZone: tzName
	"To find a time zone for Detroit:
		TimeZoneDatabase grepFor: 'Detroit'
	Then use this method to set the desired time zone."

	"TimeZoneDatabase setLocalTimeZone: 'Europe/Moscow' "
	"TimeZoneDatabase setLocalTimeZone: 'right/America/Detroit' "
	"TimeZoneDatabase setLocalTimeZone: 'posix/America/Detroit' "
	"TimeZoneDatabase setLocalTimeZone: 'America/Detroit' "
	"TimeZoneDatabase setLocalTimeZone: 'UTC' "
	"TimeZoneDatabase setLocalTimeZone: 'GMT' "
	"TimeZoneDatabase setLocalTimeZone: 'Europe/London' "
	"TimeZoneDatabase setLocalTimeZone: 'Africa/Johannesburg' "
	"TimeZoneDatabase setLocalTimeZone: 'America/Los_Angeles' "

	self systemDatabase defaultLocation: tzName.
	^ self systemDatabase defaultTimeZone
! !

!TimeZoneDatabase class methodsFor: 'private' stamp: 'dtl 1/2/2000 23:09'!
showProgressFrom: begin to: end withMessage: aString forBlock: aBlock
	"Evaluate aBlock using a progress bar display if available on this platform. The
	block receives one argument, and the argument must be a one argument block.
	If this platform does not support the progress display, use a dummy block and
	do not show the progress bar."

	self supportsProgressBar
		ifTrue: [aString displayProgressAt: Sensor cursorPoint
						from: begin
						to: end
						during: aBlock]
		ifFalse: [Transcript show: aString; cr.
				aBlock value: [ :bar | "do nothing"]]
! !

!TimeZoneDatabase class methodsFor: 'private' stamp: 'PaulDeBruicker 10/7/2011 10:39'!
supportsProgressBar
	"Answer true if this platform knows how to display a progress bar."

	^ self platformIsSqueakOrPharo ! !

!TimeZoneDatabase class methodsFor: 'instance creation' stamp: 'dtl 2/14/2007 07:53'!
systemDatabase
	"Always answer a database, even if uninitialized, in order to permit
	dependencies to be initialized. When the database is updated or replaced,
	the dependents will be notified."

	"TimeZoneDatabase systemDatabase"

	^ ThisSystemDatabase ifNil: [ThisSystemDatabase := self basicNew]! !

!TimeZoneDatabase class methodsFor: 'initialize-release' stamp: 'dtl 11/13/2004 13:50'!
thisSystemDatabase: anObject

	ThisSystemDatabase := anObject.
	self changed: #timeZoneDatabaseForSystem! !

!TimeZoneDatabase class methodsFor: 'defaulting' stamp: 'dtl 12/28/1999 20:01'!
tzExamplesPrefixPath
	"Answer the location of the timezone example database files. Assume that
	they are in the current default directory."

	^ nil

! !

!TimeZoneDatabase class methodsFor: 'private' stamp: 'dtl 1/2/2000 18:34'!
tzFilesInDirectory: aPathName inBasePath: prefixPath
	"Answer a collection of names of files in the specified directory which
	are tzfile format time zone data files. Do not search subdirectories. Do
	not include the prefixPath in the collection of answered names."

	"TimeZoneDatabase tzFilesInDirectory: 'America' inBasePath: '/usr/share/zoneinfo'"

	| fullPath dirEntries result idx relativePath |
	fullPath := self pathFromPrefix: prefixPath andName: aPathName.
	dirEntries := self directoryContents: fullPath.
	result := OrderedCollection new.
	self showProgressFrom: 1
		to: (dirEntries at: 1) size
		withMessage: ('scanning ', fullPath)
		forBlock: [ :bar |
		idx := 0.
		(dirEntries at: 1) do: [:file |
			bar value: (idx := idx + 1).
			relativePath := (self pathFromPrefix: aPathName andName: file).
			(self isTzFileType: (self pathFromPrefix: fullPath andName: file))
				ifTrue: [result add: relativePath]]].
	^ result asArray

! !

!TimeZoneDatabase class methodsFor: 'private' stamp: 'dtl 1/2/2000 18:26'!
tzFilesInDirectoryTree: aPathName
	"Search a directory recursively and answer an array of all the tzfile
	time zone files in the subtree."

	"TimeZoneDatabase tzFilesInDirectoryTree: '/usr/share/zoneinfo/'"

	| directories result idx |
	directories := OrderedCollection with: ''.
	directories addAll: (self directoryFilesIn: '' inBasePath: aPathName).
	result := OrderedCollection new.
	self showProgressFrom: 1
		to: (directories size)
		withMessage: 'selecting timezone files'
		forBlock: [ :bar |
		idx := 0.
		directories do: [:e |
			bar value: (idx := idx + 1).
			result addAll: (self tzFilesInDirectory: e inBasePath: aPathName)]].
	^ result

! !

!TimeZoneDatabase class methodsFor: 'defaulting' stamp: 'dtl 12/28/1999 20:01'!
tzPrefixPath
	"Answer the location of the timezone database files as found on a SuSE
	Linux 6.2 distribution. Modify as needed for your own system."

	^ '/usr/share/zoneinfo'

! !

!TimeZoneDatabase class methodsFor: 'version' stamp: 'dtl 2/24/2019 17:45'!
versionString
	"TimeZoneDatabase versionString"

	^ '2.1.1'! !

!TimeZoneDatabase class methodsFor: 'examples' stamp: 'dtl 2/26/2019 21:52'!
worldTimesAt: aDateAndTime
	"For the indicated time, answer a dictionary with the values of this time in each
	local time zone."

	"TimeZoneDatabase worldTimesAt: (DateAndTime fromSeconds: 0)"

	| db d |
	db := TimeZoneDatabase systemDatabase.
	db isNil ifTrue: [
		self notify: 'no system time zone database'.
		^ nil].
	d := Dictionary new.
	db allTimeZones do: [:e | d at: e timeZoneName put: (aDateAndTime inTimeZone: e)].
	^ d

! !

!TimeZoneDatabase class methodsFor: 'examples' stamp: 'dtl 2/24/2019 18:21'!
worldTimesNow
	"For the current time, answer a dictionary with the values of the current
	time in each local time zone."

	"TimeZoneDatabase worldTimesNow"

	^ TimeZoneDatabase worldTimesAt: DateAndTime now

! !

!TimeZoneDatabase methodsFor: 'time zone proxy' stamp: 'dtl 11/14/2004 13:13'!
abbreviation
	"Answer one of possibly several abbreviations for the current time zone setting"

	| abbreviations |
	abbreviations := self tzNameAbbreviations.
	^ abbreviations isEmpty
		ifTrue: []
		ifFalse: [abbreviations anyOne asString]
! !

!TimeZoneDatabase methodsFor: 'private' stamp: 'dtl 1/3/2000 23:18'!
addZone: aTimeZoneRuleSet toHierarchy: dict withKeys: keys
	"Store aTimeZoneRule in the dictionary hierarchy."

	| car cdr |
	car := (keys at: 1) asSymbol.
	(keys size == 1)
	  ifTrue: [dict at: car put: aTimeZoneRuleSet]
	  ifFalse: [
		cdr := keys copyFrom: 2 to: keys size.
		dict at: car ifAbsent: [dict at: car put: Dictionary new].
		self addZone: aTimeZoneRuleSet toHierarchy: (dict at: car) withKeys: cdr].
	^ dict

! !

!TimeZoneDatabase methodsFor: 'private' stamp: 'dtl 12/28/1999 20:01'!
addZonesTo: aCollection from: aDictionaryOfDictionaries
	"Recursively add all entries from aDictionaryOfDictionaries to aCollection."

	aDictionaryOfDictionaries do: [:e |
		(self isContainerNode: e)
			ifTrue: [self addZonesTo: aCollection from: e]
			ifFalse: [(self isTerminalLeaf: e)
				ifTrue: [aCollection add: e]
				ifFalse: [self notify: 'empty dictionary perhaps?']]]
! !

!TimeZoneDatabase methodsFor: 'querying' stamp: 'dtl 12/28/1999 20:01'!
allTimeZones
	"Answer a collection of all time zones, independent of indexing."

	"TimeZoneDatabase systemDatabase allTimeZones"

	| c |
	c := OrderedCollection new.
	self addZonesTo: c from: self timeZones.
	^ c asArray
! !

!TimeZoneDatabase methodsFor: 'normalizing' stamp: 'dtl 12/31/1999 16:24'!
combineDuplicateTables: timeZoneCollection
	"Scan timeZonecollection, combining duplicate copies of transitionTimeTable.
	If another time zone differs only in its name, then we can share transition
	time tables to avoid unnecessary duplication."

	| workingSet nextSet idx ws this that |
	workingSet := timeZoneCollection.
	nextSet := OrderedCollection new.
	idx := 0.
	self class showProgressFrom: 1 to: timeZoneCollection size withMessage: 'combining timezone rule tables'
		forBlock: [ :bar |
		[workingSet size > 1] whileTrue: [
			ws := ReadStream on: workingSet.
			this := ws next.
			bar value: idx.
			idx := idx + 1.
			[ws atEnd] whileFalse: [
				that := ws next.
				(this isEquivalentTo: that)
					ifTrue: [that transitionTimeTable: this transitionTimeTable.
							idx := idx + 1]
					ifFalse: [nextSet add: that]].
					workingSet := nextSet.
					nextSet := OrderedCollection new]]

! !

!TimeZoneDatabase methodsFor: 'accessing' stamp: 'dtl 12/30/1999 18:56'!
defaultLocation
	"If the variable has not been set, use the class default."

	defaultLocation isNil ifTrue: [^ defaultLocation := self class defaultLocation].
	^ defaultLocation! !

!TimeZoneDatabase methodsFor: 'accessing' stamp: 'dtl 11/13/2004 13:50'!
defaultLocation: aString

	"TimeZoneDatabase systemDatabase defaultLocation: 'right/America/Detroit' "

	defaultLocation := aString.
	defaultTimeZone := nil.
	self changed: #timeZoneDatabaseDefaultZone
! !

!TimeZoneDatabase methodsFor: 'accessing' stamp: 'dtl 11/9/2004 06:31'!
defaultTimeZone

	"TimeZoneDatabase systemDatabase defaultTimeZone"

	^ defaultTimeZone ifNil: [defaultTimeZone := self timeZoneFor: self defaultLocation]! !

!TimeZoneDatabase methodsFor: 'time zone proxy' stamp: 'dtl 11/13/2004 15:06'!
dstTransitionTimes

	^ self defaultTimeZone isNil
		ifTrue: [#()]
		ifFalse: [defaultTimeZone dstTransitionTimes]! !

!TimeZoneDatabase methodsFor: 'querying' stamp: 'PaulDeBruicker 3/14/2016 08:20'!
fixedTimeZoneWithOffset: anOffset atPosixSeconds: posixSeconds
	| hours hoursString tzName |
	hours := anOffset asHours asInteger.
	hoursString := hours > 0
		ifTrue: [ '+' , hours asString ]
		ifFalse: [ hours asString ].
	tzName := self gmtTimeZoneNames
		detect: [ :eachName | (eachName copyFrom: 8 to: eachName size) = hoursString ]
		ifNone: [ #UTC ].
	^ self timeZoneFor: tzName! !

!TimeZoneDatabase methodsFor: 'querying' stamp: 'PaulDeBruicker 3/13/2016 16:45'!
gmtTimeZoneNames
	^ #(#'Etc/GMT+3' #'Etc/GMT-5' #'Etc/GMT+8' #'Etc/GMT+1' #'Etc/GMT-14' #'Etc/GMT+12' #'Etc/GMT-3' #'Etc/GMT+6' #'Etc/GMT-12' #'Etc/GMT+10' #'Etc/GMT-1' #'Etc/GMT-8'  #'Etc/GMT-10' #'Etc/GMT+4' #'Etc/GMT-6'  #'Etc/GMT+9' #'Etc/GMT+2' #'Etc/GMT-4' #'Etc/GMT+7'  #'Etc/GMT-13' #'Etc/GMT+11' #'Etc/GMT-2' #'Etc/GMT-9' #'Etc/GMT+5' #'Etc/GMT-11' #'Etc/GMT-7' #'Etc/GMT-0' #'Etc/GMT+0')! !

!TimeZoneDatabase methodsFor: 'querying' stamp: 'dtl 12/28/1999 20:01'!
grepFor: aString
	"Convenience method to help find time zones for locations, especially if
	the name of the key city is known."

	"TimeZoneDatabase systemDatabase grepFor: 'Detroit' "
	"TimeZoneDatabase systemDatabase grepFor: 'etro' "
	"TimeZoneDatabase systemDatabase grepFor: 'London' "
	"TimeZoneDatabase systemDatabase grepFor: 'Perth' "
	"TimeZoneDatabase systemDatabase grepFor: 'Australia' "
	"TimeZoneDatabase systemDatabase grepFor: 'right/Australia' "
	"TimeZoneDatabase systemDatabase grepFor: 'right/Australia/Perth' "

	^ self allTimeZones select: [:e | ('*', aString, '*') match: e timeZoneName]
! !

!TimeZoneDatabase methodsFor: 'querying' stamp: 'dtl 9/13/2004 11:36'!
hoursFrom: name to: anotherTimeZoneName
	"Ask the database for the offset in hours between two time zones. The result
	may be an integer or a fraction."

	"TimeZoneDatabase systemDatabase hoursFrom: 'Europe/Amsterdam' to: 'America/Detroit'"
	"(TimeZoneDatabase systemDatabase hoursFrom: 'CET' to: 'Canada/Newfoundland') asFloat"

	^ (self offsetFrom: name to: anotherTimeZoneName) / 3600
! !

!TimeZoneDatabase methodsFor: 'adding' stamp: 'dtl 1/1/2000 14:33'!
indexAt: key put: aTimeZoneRuleSet
	"Store aTimeZoneRuleSet using hierarchical index key. Use default path separator."

	"TimeZoneDatabase new indexAt: 'right/America/Detroit' put: TimeZoneRuleSet here"

	^ self indexAt: key withSeparator: nil put: aTimeZoneRuleSet

! !

!TimeZoneDatabase methodsFor: 'adding' stamp: 'dtl 1/1/2000 14:34'!
indexAt: key withSeparator: aCharacter put: aTimeZoneRuleSet
	"Store aTimeZoneRuleSet using hierarchical index key."

	"TimeZoneDatabase new
		indexAt: 'right-America-Detroit'
		withSeparator: $-
		put: TimeZoneRuleSet here"

	| keys |
	keys := self tokensFor: key withSeparator: aCharacter.
	self addZone: aTimeZoneRuleSet toHierarchy: self timeZones withKeys: keys

! !

!TimeZoneDatabase methodsFor: 'accessing' stamp: 'dtl 12/30/1999 18:56'!
indexSeparator

	indexSeparator isNil ifTrue: [indexSeparator := self class indexSeparator].
	^ indexSeparator! !

!TimeZoneDatabase methodsFor: 'accessing' stamp: 'dtl 12/28/1999 20:01'!
indexSeparator: aString
	"Answer the substring which is used as a path separator for the time zone names."

	indexSeparator := aString
! !

!TimeZoneDatabase methodsFor: 'private' stamp: 'dtl 12/28/1999 20:01'!
isContainerNode: anObject
	"Answer true if this is the kind of object which stores other objects in the database."

	^ anObject isKindOf: Dictionary! !

!TimeZoneDatabase methodsFor: 'private' stamp: 'dtl 12/30/1999 10:03'!
isTerminalLeaf: anObject
	"Answer true if this is the kind of object which is stored as a leaf node in the database."

	^ anObject isKindOf: LocalTimeTransform! !

!TimeZoneDatabase methodsFor: 'accessing' stamp: 'dtl 12/30/1999 18:56'!
leapSecondRuleSet

	leapSecondRuleSet isNil ifTrue: [leapSecondRuleSet := LeapSecondRuleSet new].
	^ leapSecondRuleSet! !

!TimeZoneDatabase methodsFor: 'accessing' stamp: 'dtl 12/28/1999 20:01'!
leapSecondRuleSet: aLeapSecondRuleSet

	leapSecondRuleSet := aLeapSecondRuleSet! !

!TimeZoneDatabase methodsFor: 'leap seconds' stamp: 'dtl 11/13/2004 15:04'!
leapSecondsAt: aPointInTimeOrPosixSeconds
	"Answer the leap seconds for a point in time (or its integer value) for
	the active local time transform."

	^ self defaultTimeZone isNil
		ifTrue: [0]
		ifFalse: [defaultTimeZone leapSecondsAt: aPointInTimeOrPosixSeconds]! !

!TimeZoneDatabase methodsFor: 'querying' stamp: 'dtl 1/1/2000 14:36'!
leapSecondsFor: name at: aPointInTime
	"Ask the database for the offset seconds for the indicated time zone and time."

	"TimeZoneDatabase systemDatabase leapSecondsFor: 'right/America/Detroit' at: PointInTime now"
	"TimeZoneDatabase systemDatabase leapSecondsFor: 'America/Detroit' at: 944798928"
	"TimeZoneDatabase systemDatabase leapSecondsFor: 'Europe/Amsterdam' at: PointInTime now"
	"TimeZoneDatabase systemDatabase leapSecondsFor: 'America/Nowhere' at: PointInTime now"

	| tz |
	tz := (self timeZoneFor: name).
	tz isNil ifTrue: [^ nil].
	^ tz leapSecondsAt: aPointInTime! !

!TimeZoneDatabase methodsFor: 'private' stamp: 'dtl 12/30/1999 18:57'!
leapSecondsRuleFor: posixSeconds
	"Answer the leap second data record for a point in time (an Integer). Assuming
	that elements of leapSecondTable are sorted by transitionTime, this is the last
	such element for which the transition time is less than posixSeconds."

	| recordsBeforeThisTime |
	leapSecondRuleSet isNil ifTrue: [^ nil].
	leapSecondRuleSet isEmpty ifTrue: [^ nil].
	recordsBeforeThisTime := leapSecondRuleSet select: [:e | (e at: 1) < posixSeconds].
	recordsBeforeThisTime isEmpty
		ifTrue: [^ nil]
		ifFalse: [^ recordsBeforeThisTime last]! !

!TimeZoneDatabase methodsFor: 'adding' stamp: 'dtl 12/28/1999 20:01'!
loadDataFrom: aFileName

	^ self loadDataFrom: aFileName prefixPath: nil! !

!TimeZoneDatabase methodsFor: 'adding' stamp: 'dtl 12/28/1999 20:01'!
loadDataFrom: aFileName prefixPath: pathName
	"Load with condenseDuplicates true, assuming that this method is called to load an individual
	tzfile files into an existing database."

	^ self loadDataFrom: aFileName prefixPath: pathName pathSeparator: nil condenseDuplicates: true
! !

!TimeZoneDatabase methodsFor: 'adding' stamp: 'dtl 5/31/2009 12:46'!
loadDataFrom: aFileName prefixPath: pathName pathSeparator: aCharacter condenseDuplicates: aBoolean
	"If aBoolean is true, attempt to avoid duplicate copies of leap second tables
	and time zone transition tables. Set true only if loading an individual file; otherwise is
	is more efficient to use TimeZoneDatabase>>normalize after all files are loaded."

	| condense name tzArray timeZone leap matchingTimeZone |
	(aBoolean == true) ifTrue: [condense := true] ifFalse: [condense := false].	"Default to false"
	aCharacter isNil
		ifTrue: [name := aFileName]
		ifFalse: [name := self class changeSeparatorFrom: aCharacter to: self class indexSeparator in: aFileName].
	tzArray := (TzFileLoader forFile: aFileName prefixPath: pathName name: name) load.
	tzArray isNil ifTrue: [^ nil].
	timeZone := tzArray at: 1.
	leap := tzArray at: 2.
	"Notes: tzArray at: 3 is a possibly nil TZ environment string, from a version
	2 tzfile, currently unused. tzArray at: 4 is a possibly nil copy of version 1
	format data (if parsing a version 2 file) available for debugging."

	self indexAt: timeZone timeZoneName put: timeZone.
	"Normally, all time zones which honor leap second rules can share a
	common LeapSecondRuleSet."
	condense ifTrue: [
		leap isNil ifFalse: [
			self updateLeapSecondRulesFrom: leap.
			(self leapSecondRuleSet = leap) ifTrue: [
				timeZone leapSecondRuleSet: leap]]].
	"If another time zone differs only in its name, then we can share transition
	time tables to avoid unnecessary duplication."
	condense ifTrue: [
		matchingTimeZone := self allTimeZones detect: [:e | timeZone isEquivalentTo: e] ifNone: [nil].
		matchingTimeZone isNil ifFalse: [timeZone transitionTimeTable: matchingTimeZone transitionTimeTable]]
! !

!TimeZoneDatabase methodsFor: 'offsets from UTC' stamp: 'dtl 11/13/2004 15:05'!
localOffsetSecondsAt: aPointInTimeOrPosixSeconds
	"Answer the offset seconds for a point in time (or its integer value) for
	the active local time transform."

	^ self defaultTimeZone isNil
		ifTrue: [0]
		ifFalse: [defaultTimeZone localOffsetSecondsAt: aPointInTimeOrPosixSeconds]
! !

!TimeZoneDatabase methodsFor: 'time zone proxy' stamp: 'dtl 11/13/2004 22:43'!
name
	"The name of the active time zone, when the database is used as a TimeZone."

	^ (self defaultLocation ifNil: [super name]) asString! !

!TimeZoneDatabase methodsFor: 'normalizing' stamp: 'dtl 12/28/1999 20:01'!
normalize
	"Eliminate duplication of rule sets where possible."

	| allEntries |
	allEntries := self allTimeZones.
	self setLeapSecondRulesFrom: allEntries.
	self combineDuplicateTables: allEntries! !

!TimeZoneDatabase methodsFor: 'time zone proxy' stamp: 'dtl 2/26/2019 22:43'!
offset

	^ Duration seconds: (self localOffsetSecondsAt: DateAndTime now)! !

!TimeZoneDatabase methodsFor: 'querying' stamp: 'dtl 12/30/1999 18:58'!
offsetFor: name at: aPointInTime
	"Ask the database for the offset seconds for the indicated time zone and time."

	"TimeZoneDatabase systemDatabase offsetFor: 'America/Detroit' at: 944798928"
	"TimeZoneDatabase systemDatabase offsetFor: 'Europe/Amsterdam' at: PointInTime now"
	"TimeZoneDatabase systemDatabase offsetFor: 'America/Nowhere' at: PointInTime now"

	| tz |
	tz := (self timeZoneFor: name).
	tz isNil ifTrue: [^ nil].
	^ tz localOffsetSecondsAt: aPointInTime! !

!TimeZoneDatabase methodsFor: 'querying' stamp: 'dtl 2/26/2019 22:44'!
offsetFrom: name to: anotherTimeZoneName
	"Ask the database for the current offset in seconds between two time zones."

	"TimeZoneDatabase systemDatabase offsetFrom: 'Europe/Amsterdam' to: 'America/Detroit'"

	| now |
	now := DateAndTime now.
	^ (self offsetFor: name at: now) - (self offsetFor: anotherTimeZoneName at: now)
! !

!TimeZoneDatabase methodsFor: 'querying' stamp: 'PaulDeBruicker 3/14/2016 08:10'!
popularTimeZoneNames
	^ #(#'Pacific/Auckland' #'Australia/Sydney' #'Asia/Tokyo' #'Australia/Perth' #'Asia/Hong_Kong' #'Asia/Calcutta' #'Europe/Moscow' #'Asia/Jerusalem' #'Europe/Amsterdam' #'Europe/London' #UTC #'America/Sao_Paulo' #'America/Argentina/Buenos_Aires' #'America/New_York' #'America/Chicago' #'America/Denver' #'America/Los_Angeles' #'Pacific/Honolulu')! !

!TimeZoneDatabase methodsFor: 'normalizing' stamp: 'dtl 12/28/1999 20:01'!
setLeapSecondRulesFrom: timeZoneCollection
	"Scan timeZoneCollection looking for leap second rule sets. Replace the
	individual leap second rule sets with a single common LeapSecondRuleSet
	which I hold on behalf of all the time zones. For time zones with no leap
	second compensation, ensure that leapSecondRuleSet is nil."

	| notEmpty |
	notEmpty := [:e | (e leapSecondRuleSet isNil) or: [e leapSecondRuleSet isEmpty]].
	(timeZoneCollection reject: notEmpty) do: [:e |
		self updateLeapSecondRulesFrom: e leapSecondRuleSet.
		e leapSecondRuleSet: self leapSecondRuleSet].
	(timeZoneCollection select: notEmpty) do: [:e |
		e leapSecondRuleSet: nil]




! !

!TimeZoneDatabase methodsFor: 'querying' stamp: 'dtl 12/28/1999 20:01'!
timeZoneFor: name

	"TimeZoneDatabase systemDatabase timeZoneFor: 'America/Detroit' "

	^ self timeZoneFor: name pathSeparator: nil
! !

!TimeZoneDatabase methodsFor: 'querying' stamp: 'dtl 12/30/1999 18:34'!
timeZoneFor: name pathSeparator: aCharacter
	"Allow the default path separator to be overridden for queries."

	"TimeZoneDatabase systemDatabase timeZoneFor: 'America/Detroit' pathSeparator: $/ "
	"TimeZoneDatabase systemDatabase timeZoneFor: 'America%Detroit' pathSeparator: $% "

	| keys |
	keys := self tokensFor: name withSeparator: aCharacter.
	^ self timeZoneIndexedBy: keys inHierarchy: self timeZones! !

!TimeZoneDatabase methodsFor: 'private' stamp: 'dtl 12/28/1999 20:01'!
timeZoneIndexedBy: anArrayOfKeys inHierarchy: aDictionaryTree
	"Answer the time zone indexed by anArrayOfKeys, or nil if not found."

	| car cdr entry |
	car := (anArrayOfKeys at: 1) asSymbol.
	entry := aDictionaryTree at: car ifAbsent: [^ nil].
	(anArrayOfKeys size == 1)
		ifTrue: [(self isTerminalLeaf: entry) ifTrue: [^ entry] ifFalse: [^ nil]]
		ifFalse: [(self isContainerNode: entry)
					ifFalse: [^ nil]
					ifTrue: [cdr := anArrayOfKeys copyFrom: 2 to: anArrayOfKeys size.
							^ self timeZoneIndexedBy: cdr inHierarchy: entry]]
! !

!TimeZoneDatabase methodsFor: 'querying' stamp: 'dtl 11/9/2004 06:48'!
timeZoneName
	"Answer the name of the active local time transform"

	^ self defaultLocation asString
! !

!TimeZoneDatabase methodsFor: 'querying' stamp: 'dtl 12/28/1999 20:01'!
timeZoneNames

	"TimeZoneDatabase systemDatabase timeZoneNames"

	^ self allTimeZones collect: [:e | e timeZoneName]

! !

!TimeZoneDatabase methodsFor: 'querying' stamp: 'PaulDeBruicker 3/14/2016 08:16'!
timeZoneWithOffset: anOffset atPosixSeconds: posixSeconds
	self popularTimeZoneNames
		do: [ :eachName | 
			| tz |
			tz := self timeZoneFor: eachName.
			(tz localOffsetSecondsAt: posixSeconds) = anOffset asSeconds
				ifTrue: [ ^ tz ] ].
	^ self fixedTimeZoneWithOffset: anOffset atPosixSeconds: posixSeconds! !

!TimeZoneDatabase methodsFor: 'accessing' stamp: 'dtl 12/30/1999 18:58'!
timeZones

	timeZones isNil ifTrue: [timeZones := Dictionary new].
	^ timeZones! !

!TimeZoneDatabase methodsFor: 'private' stamp: 'dtl 12/30/1999 18:51'!
tokensFor: aString withSeparator: aCharacter
	"Answer an array of tokens from aString separated by aCharacter.
	Note: String>>findTokens is not used because some Smalltalks do not have it."

	"TimeZoneDatabase new tokensFor: 'right-America-Detroit' withSeparator: $- "

	| c s sep |
	c  := OrderedCollection new.
	s := ReadStream on: aString.
	aCharacter isNil
		ifTrue: [sep := self indexSeparator]
		ifFalse: [sep := aCharacter].
	[s atEnd] whileFalse: [
		[s peek == sep] whileTrue: [s next].
		c add: (s upTo: sep)].
	^ c asArray



! !

!TimeZoneDatabase methodsFor: 'querying' stamp: 'dtl 11/14/2004 13:11'!
tzNameAbbreviations
	"Answer the abbreviations for the active local time transform"

	| zone |
	zone := self defaultTimeZone.
	^ zone isNil
		ifTrue: [#()]
		ifFalse: [zone tzNameAbbreviations]
! !

!TimeZoneDatabase methodsFor: 'adding' stamp: 'dtl 12/28/1999 20:01'!
updateLeapSecondRulesFrom: aLeapSecondRuleSet
	"If my leapSecondRuleSet is nil or empty, update it with aLeapSecondRuleSet.
	Normally, all leap second rule sets will be identical system-wide. Therefore,
	if my leapSecondRuleSet is not empty, and aLeapSecondRuleSet is also not
	empty, check to see if they are equal. If not, warn the user."

	((leapSecondRuleSet isNil) or: [leapSecondRuleSet isEmpty])
		ifTrue: [self leapSecondRuleSet: aLeapSecondRuleSet]
		ifFalse: [aLeapSecondRuleSet isEmpty not
			ifTrue: [(leapSecondRuleSet leapSecondTable ~= aLeapSecondRuleSet leapSecondTable)
				ifTrue: [self notify: ('Leap second rules differ. ',
					'This may OK if you are updating with newer tzfile files.').
					self leapSecondRuleSet: aLeapSecondRuleSet]]]! !

Object subclass: #TimeZoneRule
	instanceVariableNames: 'transitionTime offsetSeconds isDstFlag abbreviation isStdTimeTransition isUtcTransition'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'TZ-Olson'!

!TimeZoneRule commentStamp: '<historical>' prior: 0!
I represent a record of transition time data in a ZoneRuleSet. One such
record is in effect at any point in time in a given time zone, providing the
means to calculate seconds offset from UTC.!

!TimeZoneRule methodsFor: 'comparing' stamp: 'dtl 12/28/1999 20:01'!
= aTimeZoneRule
	"Assumes that transitionTime is a LargePositiveInteger, offsetSeconds is a SmallInteger,
	abbreviation is a Symbol."

	^ (self species == aTimeZoneRule species) and: [
		(offsetSeconds == aTimeZoneRule offsetSeconds) and: [
		  (abbreviation == aTimeZoneRule abbreviation) and: [
			(transitionTime = aTimeZoneRule transitionTime) and: [
			  (isDstFlag == aTimeZoneRule isDstFlag) and: [
				(isStdTimeTransition == aTimeZoneRule isStdTimeTransition) and: [
				  (isUtcTransition == aTimeZoneRule isUtcTransition)]]]]]]
! !

!TimeZoneRule methodsFor: 'accessing' stamp: 'dtl 12/28/1999 20:01'!
abbreviation

	^ abbreviation! !

!TimeZoneRule methodsFor: 'accessing' stamp: 'dtl 12/28/1999 20:01'!
abbreviation: aString

	abbreviation := aString! !

!TimeZoneRule methodsFor: 'comparing' stamp: 'dtl 12/28/1999 20:01'!
hash
	"Hash is reimplemented because = is implemented."

	^ (((((transitionTime hash
		+ offsetSeconds hash)
		+ isDstFlag hash)
		+ abbreviation hash)
		+ isStdTimeTransition hash)
		+ isUtcTransition hash)
! !

!TimeZoneRule methodsFor: 'accessing' stamp: 'dtl 12/28/1999 20:01'!
isDstFlag

	^ isDstFlag! !

!TimeZoneRule methodsFor: 'accessing' stamp: 'dtl 12/28/1999 20:01'!
isDstFlag: anInteger

	isDstFlag := anInteger! !

!TimeZoneRule methodsFor: 'accessing' stamp: 'dtl 12/28/1999 20:01'!
isStdTimeTransition

	^ isStdTimeTransition! !

!TimeZoneRule methodsFor: 'accessing' stamp: 'dtl 12/28/1999 20:01'!
isStdTimeTransition: aBoolean

	isStdTimeTransition := aBoolean! !

!TimeZoneRule methodsFor: 'accessing' stamp: 'dtl 12/28/1999 20:01'!
isUtcTransition

	^ isUtcTransition! !

!TimeZoneRule methodsFor: 'accessing' stamp: 'dtl 12/28/1999 20:01'!
isUtcTransition: aBoolean

	isUtcTransition := aBoolean! !

!TimeZoneRule methodsFor: 'accessing' stamp: 'dtl 12/28/1999 20:01'!
offsetSeconds

	^ offsetSeconds! !

!TimeZoneRule methodsFor: 'accessing' stamp: 'dtl 12/28/1999 20:01'!
offsetSeconds: anInteger

	offsetSeconds := anInteger! !

!TimeZoneRule methodsFor: 'printing' stamp: 'dtl 11/20/2004 12:47'!
printOn: aStream

	"^ self printSimplyOn: aStream"
	^ self printVerboseOn: aStream
! !

!TimeZoneRule methodsFor: 'printing' stamp: 'dtl 12/28/1999 20:01'!
printSimplyOn: aStream

	super printOn: aStream.
	aStream nextPutAll: ' ('.
	transitionTime printOn: aStream.
	aStream nextPutAll: ', '.
	offsetSeconds printOn: aStream.
	aStream nextPutAll: ', '; nextPutAll: abbreviation; nextPut: $)! !

!TimeZoneRule methodsFor: 'printing' stamp: 'dtl 2/26/2019 22:53'!
printVerboseOn: aStream
	"Print enough detail to allow the output to be compared with the output of
	the zdump test utility, as found on may Unix-like systems."

	(DateAndTime fromUnixTime: self transitionTime) printOn: aStream.
	aStream nextPutAll: (' ', self abbreviation, ' (').
	transitionTime printOn: aStream.
	aStream nextPutAll: ', '.
	offsetSeconds printOn: aStream.
	aStream nextPutAll: ', '.
	isDstFlag printOn: aStream.
	aStream nextPutAll: ', '.
	abbreviation printOn: aStream.
	aStream nextPutAll: ', '.
	isStdTimeTransition printOn: aStream.
	aStream nextPutAll: ', '.
	isUtcTransition printOn: aStream.
	aStream nextPut: $)! !

!TimeZoneRule methodsFor: 'printing' stamp: 'dtl 2/26/2019 23:03'!
printVerboseOn: aStream withLeapSeconds: anInteger
	"Print enough detail to allow the output to be compared with the output of
	the zdump test utility, as found on may Unix-like systems. This rule does
	not know about leap seconds, which are maintained in a separate rule table,
	so any leap second compensation must be provided as a parameter."

	| dt dtUtc leap timeWithLeapOffset |
	anInteger isNil ifTrue: [leap := 0] ifFalse: [leap := anInteger].
	timeWithLeapOffset := self transitionTime - leap.
	dt := DateAndTime fromUnixTime: timeWithLeapOffset.
	dtUtc := DateAndTime fromUnixTime: (timeWithLeapOffset + self offsetSeconds).
	(dt at: 1) printOn: aStream.
	aStream nextPut: $ .
	(dt at: 2) printOn: aStream.
	aStream nextPutAll: ' UTC = '.
	(dtUtc at: 1) printOn: aStream.
	aStream nextPut: $ .
	(dtUtc at: 2) printOn: aStream.
	aStream nextPutAll: (' ', self abbreviation, ' (').
	transitionTime printOn: aStream.
	aStream nextPutAll: ', '.
	offsetSeconds printOn: aStream.
	aStream nextPutAll: ', '.
	isDstFlag printOn: aStream.
	aStream nextPutAll: ', '.
	abbreviation printOn: aStream.
	aStream nextPutAll: ', '.
	isStdTimeTransition printOn: aStream.
	aStream nextPutAll: ', '.
	isUtcTransition printOn: aStream.
	aStream nextPut: $)! !

!TimeZoneRule methodsFor: 'accessing' stamp: 'dtl 12/28/1999 20:01'!
transitionTime

	^ transitionTime! !

!TimeZoneRule methodsFor: 'accessing' stamp: 'dtl 12/28/1999 20:01'!
transitionTime: anInteger

	transitionTime := anInteger! !

Object subclass: #TzFileLoader
	instanceVariableNames: 'name fileStream tzNames transitionTable leapSeconds'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'TZ-Olson'!

!TzFileLoader commentStamp: '<historical>' prior: 0!
I understand the format of a tzfile(5) timezone database file. I know
how to load data from a tzfile to create instances of TimeZoneRuleSet
and LeapSecondRuleSet.

A tzfile database file is typically generated by the zic(1) compiler
distributed as part of the public domain timezone database in the
~ftp/pub directory of elsie.nci.nih.gov FTP server.
!

!TzFileLoader class methodsFor: 'instance creation' stamp: 'PaulDeBruicker 3/13/2016 22:39'!
forFile: aTzFileName prefixPath: aString
	"Load an instance of TimeZoneRuleSet from a tzfile file. Set the time zone name to be
	the same as the file name."

	"(TzFileLoader forFile: 'right/America/Detroit' prefixPath: '/usr/share/zoneinfo/') load"

	"(TzFileLoader forFile: 'right-America-Detroit' prefixPath: nil) load"

	"(TzFileLoader forFile: 'America/Barbados' prefixPath: '/usr/share/zoneinfo/') load"

	"(TzFileLoader forFile: 'America/Chicago' prefixPath: '/usr/share/zoneinfo/') load"

	"(TzFileLoader forFile: 'America/Goose:=Bay' prefixPath: '/usr/share/zoneinfo/') load"

	"(TzFileLoader forFile: 'Europe/Zurich' prefixPath: '/usr/share/zoneinfo/') load"

	"(TzFileLoader forFile: 'Etc/GMT-7' prefixPath: '/usr/share/zoneinfo/') load"

	^ self forFile: aTzFileName prefixPath: aString name: aTzFileName! !

!TzFileLoader class methodsFor: 'instance creation' stamp: 'dtl 12/31/1999 18:13'!
forFile: aTzFileName prefixPath: aString name: aName
	"Load an instance of TimeZoneRuleSet from a tzfile file."

	"(TzFileLoader forFile: 'right/America/Detroit' prefixPath: '/usr/share/zoneinfo/') load"
	"(TzFileLoader forFile: 'right-America-Detroit' prefixPath: nil name: 'right/America/Detroit') load"


	| path fileStream |

	aString isNil
		ifTrue: [path := aTzFileName]
		ifFalse: [path := aString, aTzFileName].
	fileStream := self readStreamFor: path.
	^ super new fileStream: fileStream; name: aName asSymbol
! !

!TzFileLoader class methodsFor: 'private' stamp: 'dtl 12/31/1999 16:50'!
platform
	"Answer what version of Smalltalk we are using."

	^ TimeZoneDatabase platform! !

!TzFileLoader class methodsFor: 'private' stamp: 'PaulDeBruicker 10/7/2011 10:35'!
platformIsSqueakOrPharo
	^self platform == #Squeak or: [self platform == #Pharo]! !

!TzFileLoader class methodsFor: 'private' stamp: 'PaulDeBruicker 10/7/2011 10:35'!
readStreamFor: aFileName
	"Answer a binary FileStream for reading a tzfile data file."

	"TzFileLoader readStreamFor: '/tmp'"

	(self platformIsSqueakOrPharo ) ifTrue: [
		^ ((Smalltalk at: #FileStream) readOnlyFileNamed: aFileName) binary].
	(self platform == #VisualWorks) ifTrue: [
		^ ((Smalltalk at: #Filename) named: aFileName) readStream binary].
	(self platform == #'Smalltalk/X') ifTrue: [ | s |
		s := ((Smalltalk at: #Filename) named: aFileName) readStream.
		s isNil ifTrue: [^ nil] ifFalse: [^ s binary]]

! !

!TzFileLoader methodsFor: 'accessing' stamp: 'dtl 12/28/1999 20:01'!
fileStream

	^ fileStream! !

!TzFileLoader methodsFor: 'accessing' stamp: 'dtl 12/28/1999 20:01'!
fileStream: aFileStream

	fileStream := aFileStream! !

!TzFileLoader methodsFor: 'tzfile decoding' stamp: 'dtl 1/1/2000 13:59'!
getTzNameAbbreviationsFrom: aString
	"Read null-terminated time zone name abbreviations"

	| names strm |
	names := OrderedCollection new.
	strm := ReadStream on: aString.
	[strm atEnd] whileFalse: [names add: ((strm upTo: 0) asString asSymbol)].
	^ names
! !

!TzFileLoader methodsFor: 'tzfile decoding' stamp: 'dtl 5/31/2009 12:56'!
leapSecondStruct64From: aStream

	^ Array with: (self nextInt64FromNetworkOrderedBytes: aStream)
			with: (self nextIntFromNetworkOrderedBytes: aStream)
! !

!TzFileLoader methodsFor: 'tzfile decoding' stamp: 'dtl 12/28/1999 20:01'!
leapSecondStructFrom: aStream

	^ Array with: (self nextIntFromNetworkOrderedBytes: aStream)
			with: (self nextIntFromNetworkOrderedBytes: aStream)
! !

!TzFileLoader methodsFor: 'tzfile loading' stamp: 'pad 2/28/2016 21:49'!
load
	"Load from my fileStream, and answer an array of a TimeZoneRuleSet, a
	LeapSecondRuleSet, and a possibly nil copy of tzfile format 1 data (non-nil
	only when parsing a version 2 file). The name and fileStream instance variables
	should already be set. Close fileStream when the file load is complete."

	| tzfileVersion result |

	"Check magic ID for the file"
	((self magic: fileStream) = 'TZif') ifFalse: [
		self notify: 'bad magic for zoneinfo file'.
		fileStream close.
		^ nil].

	"The tzfile format version, either null or $2. Version 2 format is a backward
	compatible extension."
	tzfileVersion := fileStream next.

	"Skip remaining 15 unused characters"
	fileStream next: 15.

	tzfileVersion = 0
		ifTrue: [result := self loadVersion1tzfile]
		ifFalse: [tzfileVersion = $2 asciiValue
				ifTrue: [result := self loadVersion2tzfile]
				ifFalse: [tzfileVersion = $3 asciiValue ifTrue:[result:=self loadVersion3tzfile] ifFalse:[
					self notify: 'unsupported tzfile format version ', (Character value:tzfileVersion) asString.
					"assume backward compatibility, use version 1 file format"
					result := self loadVersion1tzfile]]].

	fileStream atEnd ifFalse: [
		self notify: 'zoneinfo parse error, should be at end of file'.
		fileStream close.
		^ nil].
	fileStream close.
	
	^ result
! !

!TzFileLoader methodsFor: 'tzfile loading' stamp: 'PaulDeBruicker 3/14/2016 08:06'!
loadVersion1tzfile
	"This is the original tzfile format. Load remaining data from fileStream."

	| ttisgmtcnt ttisstdcnt leapcnt timecnt typecnt charcnt transitionTimes ttinfoIndices ttinfoStructs abbreviations isStd isGmt rec timeZoneRuleSet leapSecondRuleSet |
	"Read header information"
	ttisgmtcnt := self nextUnsignedIntFromNetworkOrderedBytes: fileStream.
	ttisstdcnt := self nextUnsignedIntFromNetworkOrderedBytes: fileStream.
	leapcnt := self nextUnsignedIntFromNetworkOrderedBytes: fileStream.
	timecnt := self nextUnsignedIntFromNetworkOrderedBytes: fileStream.
	typecnt := self nextUnsignedIntFromNetworkOrderedBytes: fileStream.
	charcnt := self nextUnsignedIntFromNetworkOrderedBytes: fileStream.	"Read data for transition times"
	transitionTimes := OrderedCollection new.
	timecnt timesRepeat: [ transitionTimes add: (self nextIntFromNetworkOrderedBytes: fileStream) ].
	ttinfoIndices := OrderedCollection new.
	timecnt timesRepeat: [ ttinfoIndices add: (self nextUnsignedCharFrom: fileStream) ].	"Read ttinfo data structures"
	ttinfoStructs := OrderedCollection new.
	typecnt timesRepeat: [ ttinfoStructs add: (self ttinfoStructFrom: fileStream) ].	"Read time zone name abbreviations"
	fileStream atEnd
		ifTrue: [ self error: 'unexpected end of file' ].
	abbreviations := fileStream next: charcnt.	"Note: tzNames is redundant, as the same information will be saved in the transitionTimeTable.
	The names are stored separately here as a convenience for browsing time zones."
	tzNames := self getTzNameAbbreviationsFrom: abbreviations.	"Read leap second information"
	leapSeconds := OrderedCollection new.
	leapcnt timesRepeat: [ leapSeconds add: (self leapSecondStructFrom: fileStream) ].	"Flags for standard/wall indicators, to indicate if transition times associated with
	local time types  were specified as standard  time or wall clock time, used
	when a time zone file is used in handling POSIX-style time zone environment
	variables."
	isStd := OrderedCollection new: typecnt.
	ttisstdcnt timesRepeat: [ isStd add: (self nextBooleanCharFrom: fileStream) ].	"Store the flags as element 4 of the corresponding ttinfoStruct"
	(1 to: ttisstdcnt) do: [ :e | (ttinfoStructs at: e) at: 4 put: (isStd at: e) ].	"Flags for UTC/local indicators, tell whether the transition times associated with
	local time types were specified as UTC or local time, used when a time zone file
	is used in handling POSIX-style time zone environment variables."
	isGmt := OrderedCollection new: typecnt.
	ttisgmtcnt timesRepeat: [ isGmt add: (self nextBooleanCharFrom: fileStream) ].	"Store the flags as element 5 of the corresponding ttinfoStruct"
	(1 to: ttisstdcnt) do: [ :e | (ttinfoStructs at: e) at: 5 put: (isGmt at: e) ].	"Relate the transition times to the ttinfoStruct data"
	transitionTable := Array new: timecnt.
	(1 to: timecnt)
		do: [ :idx | 
			rec := ttinfoStructs at: (ttinfoIndices at: idx) + 1.
			transitionTable
				at: idx
				put:
					(TimeZoneRule new
						transitionTime: (transitionTimes at: idx);
						offsetSeconds: (rec at: 1);
						isDstFlag: (rec at: 2);
						abbreviation:
								(((ReadStream on: abbreviations)
										next: (rec at: 3);
										yourself) upTo: 0) asString asSymbol;
						isStdTimeTransition: (rec at: 4);
						isUtcTransition: (rec at: 5)) ].	"Create an instance of LeapSecondRuleSet, or nil if no rules were found in this tzfile."
	leapSeconds size = 0
		ifTrue: [ leapSecondRuleSet := nil ]
		ifFalse: [ leapSecondRuleSet := LeapSecondRuleSet new leapSecondTable: leapSeconds ].	"Create an instance of TimeZoneRuleSet"
	('Etc/GMT*' match: name)
		ifTrue: [ 
			"make a FixedTimeZoneRuleSet for a time zone without transitions e.g. Etc/GMT-7"
			timeZoneRuleSet := FixedTimeZoneRuleSet new.
			timeZoneRuleSet offset: ttinfoStructs first first negated ]
		ifFalse: [ 
			"make a TimeZoneRuleSet for a time zone with transitions"
			timeZoneRuleSet := TimeZoneRuleSet new ].
	timeZoneRuleSet timeZoneName: name.
	timeZoneRuleSet transitionTimeTable: transitionTable.
	timeZoneRuleSet leapSecondRuleSet: leapSecondRuleSet.
	timeZoneRuleSet tzNameAbbreviations: tzNames asArray.
	^ Array with: timeZoneRuleSet with: leapSecondRuleSet with: nil! !

!TzFileLoader methodsFor: 'tzfile loading' stamp: 'dtl 5/31/2009 13:12'!
loadVersion2tzfile
	"This is version 2 tzfile format. Load remaining data from fileStream."

	| ttisgmtcnt ttisstdcnt leapcnt timecnt typecnt charcnt
	transitionTimes ttinfoIndices ttinfoStructs abbreviations isStd isGmt rec
	timeZoneRuleSet leapSecondRuleSet posixTzEnvString versionOneData |

	"Read the version 1 compatible initial section, ignoring the results"
	versionOneData := self loadVersion1tzfile.

	"From the tzfile man page:
	For version-2-format time zone files, the above header and data is followed by a second
	header and data, identical in format except that eight bytes are used for each transition
	time or leap second time.  After the second header and data comes a newline-enclosed,
	POSIX-TZ-environment-variable-style string for use in handling instants after the last
	transition time stored in the file (with nothing between the newlines if there is no
	POSIX representation for such instants)."

	"Read header information (this is a second copy of the file header, and can be ignored)"
		"Check magic ID for the file"
	((self magic: fileStream) = 'TZif') ifFalse: [
		self notify: 'bad magic for zoneinfo file'.
		^ nil].

	"The tzfile format version, either null or $2. Version 2 format is a backward
	compatible extension."
	fileStream next = $2 asciiValue ifFalse: [
		self notify: 'expected version number 2 in embedded header portion of file'.
		^ nil].

	"Skip remaining 15 unused characters"
	fileStream next: 15.

	ttisgmtcnt := self nextUnsignedIntFromNetworkOrderedBytes: fileStream.
	ttisstdcnt := self nextUnsignedIntFromNetworkOrderedBytes: fileStream.
	leapcnt := self nextUnsignedIntFromNetworkOrderedBytes: fileStream.
	timecnt := self nextUnsignedIntFromNetworkOrderedBytes: fileStream.
	typecnt := self nextUnsignedIntFromNetworkOrderedBytes: fileStream.
	charcnt := self nextUnsignedIntFromNetworkOrderedBytes: fileStream.

	"Read data for transition times"
	transitionTimes := OrderedCollection new.
	timecnt timesRepeat: [transitionTimes add: (self nextInt64FromNetworkOrderedBytes: fileStream)].
	ttinfoIndices := OrderedCollection new.
	timecnt timesRepeat: [ttinfoIndices add: (self nextUnsignedCharFrom: fileStream)].

	"Read ttinfo data structures"
	ttinfoStructs := OrderedCollection new.
	typecnt timesRepeat: [ttinfoStructs add: (self ttinfoStructFrom: fileStream)].
	"typecnt timesRepeat: [ttinfoStructs add: (self ttinfoStruct64From: fileStream)]."

	"Read time zone name abbreviations"
	fileStream atEnd ifTrue: [self error: 'unexpected end of file'].
	abbreviations := fileStream next: charcnt.
	"Note: tzNames is redundant, as the same information will be saved in the transitionTimeTable.
	The names are stored separately here as a convenience for browsing time zones."
	tzNames := self getTzNameAbbreviationsFrom: abbreviations.

	"Read leap second information"
	leapSeconds := OrderedCollection new.
	leapcnt timesRepeat: [leapSeconds add: (self leapSecondStruct64From: fileStream)].

	"Flags for standard/wall indicators, to indicate if transition times associated with
	local time types  were specified as standard  time or wall clock time, used
	when a time zone file is used in handling POSIX-style time zone environment
	variables."
	isStd := OrderedCollection new: typecnt.
	ttisstdcnt timesRepeat: [isStd add: (self nextBooleanCharFrom: fileStream)].
	"Store the flags as element 4 of the corresponding ttinfoStruct"
	(1 to: ttisstdcnt) do: [:e | (ttinfoStructs at: e) at: 4 put: (isStd at: e)].

	"Flags for UTC/local indicators, tell whether the transition times associated with
	local time types were specified as UTC or local time, used when a time zone file
	is used in handling POSIX-style time zone environment variables."
	isGmt := OrderedCollection new: typecnt.
	ttisgmtcnt timesRepeat: [isGmt add: (self nextBooleanCharFrom: fileStream)].
	"Store the flags as element 5 of the corresponding ttinfoStruct"
	(1 to: ttisstdcnt) do: [:e | (ttinfoStructs at: e) at: 5 put: (isGmt at: e)].

	"Read a newline-enclosed, POSIX-TZ-environment-variable-style string for use in
	handling instants after the last transition time stored in the file (an empty string
	if there is no POSIX representation for such instants)."
	posixTzEnvString := fileStream upToEnd.
	posixTzEnvString
		ifNotNil: [(posixTzEnvString first = Character lf asciiValue
					and: [posixTzEnvString last = Character lf asciiValue])
				ifFalse: [self notify: 'expected TZ environment string surrounded by Character lf'.
					^ nil].
			posixTzEnvString := posixTzEnvString asString allButFirst allButLast].

	"Relate the transition times to the ttinfoStruct data"
	transitionTable := Array new: timecnt.
	(1 to: timecnt) do: [:idx |
		rec := ttinfoStructs at: ((ttinfoIndices at: idx) + 1).
		transitionTable at: idx put:
			(TimeZoneRule new
				transitionTime: (transitionTimes at: idx);
				offsetSeconds: (rec at: 1);
				isDstFlag: (rec at: 2);
				abbreviation: (((ReadStream on: abbreviations)
								next: (rec at: 3); yourself) upTo: 0) asString asSymbol;
				isStdTimeTransition: (rec at: 4);
				isUtcTransition: (rec at: 5))].

	"Create an instance of LeapSecondRuleSet, or nil if no rules were found in this tzfile."
	(leapSeconds size = 0)
		ifTrue: [leapSecondRuleSet := nil]
		ifFalse: [leapSecondRuleSet := LeapSecondRuleSet new leapSecondTable: leapSeconds].

	"Create an instance of TimeZoneRuleSet"
	timeZoneRuleSet := TimeZoneRuleSet new.
	timeZoneRuleSet timeZoneName: name.
	timeZoneRuleSet transitionTimeTable: transitionTable.
	timeZoneRuleSet leapSecondRuleSet: leapSecondRuleSet.
	timeZoneRuleSet tzNameAbbreviations: tzNames asArray.
	timeZoneRuleSet posixTzEnvString: posixTzEnvString.
	^ Array with: timeZoneRuleSet
		with: leapSecondRuleSet
		with: versionOneData
! !

!TzFileLoader methodsFor: 'tzfile loading' stamp: 'pad 2/28/2016 21:39'!
loadVersion3tzfile
	"This is version 3 tzfile format. Load remaining data from fileStream."

	| ttisgmtcnt ttisstdcnt leapcnt timecnt typecnt charcnt
	transitionTimes ttinfoIndices ttinfoStructs abbreviations isStd isGmt rec
	timeZoneRuleSet leapSecondRuleSet posixTzEnvString versionOneData |

	"Read the version 1 compatible initial section, ignoring the results"
	versionOneData := self loadVersion1tzfile.

	"From the tzfile man page:
	For version-2-format time zone files, the above header and data is followed by a second
	header and data, identical in format except that eight bytes are used for each transition
	time or leap second time.  After the second header and data comes a newline-enclosed,
	POSIX-TZ-environment-variable-style string for use in handling instants after the last
	transition time stored in the file (with nothing between the newlines if there is no
	POSIX representation for such instants)."
     
	"The version 3 format time zone file, the hours part of the transition times may be signed and range 	from -167 through 167 instead of the POSIX values of 0 to 24. Second, DST is in effect all year if it 	starts January 1 at 00:00 and ends December 31 at 24:00 plus the differnece betwwen daylight saving 	and standard time."

	"Read header information (this is a second copy of the file header, and can be ignored)"
		"Check magic ID for the file"
	((self magic: fileStream) = 'TZif') ifFalse: [
		self notify: 'bad magic for zoneinfo file'.
		^ nil].

	"The tzfile format version, either null or $2. Version 2 format is a backward
	compatible extension."
	fileStream next = $3 asciiValue ifFalse: [
		self notify: 'expected version number 3 in embedded header portion of file'.
		^ nil].

	"Skip remaining 15 unused characters"
	fileStream next: 15.

	ttisgmtcnt := self nextUnsignedIntFromNetworkOrderedBytes: fileStream.
	ttisstdcnt := self nextUnsignedIntFromNetworkOrderedBytes: fileStream.
	leapcnt := self nextUnsignedIntFromNetworkOrderedBytes: fileStream.
	timecnt := self nextUnsignedIntFromNetworkOrderedBytes: fileStream.
	typecnt := self nextUnsignedIntFromNetworkOrderedBytes: fileStream.
	charcnt := self nextUnsignedIntFromNetworkOrderedBytes: fileStream.

	"Read data for transition times"
	transitionTimes := OrderedCollection new.
	timecnt timesRepeat: [transitionTimes add: (self nextInt64FromNetworkOrderedBytes: fileStream)].
	ttinfoIndices := OrderedCollection new.
	timecnt timesRepeat: [ttinfoIndices add: (self nextUnsignedCharFrom: fileStream)].

	"Read ttinfo data structures"
	ttinfoStructs := OrderedCollection new.
	typecnt timesRepeat: [ttinfoStructs add: (self ttinfoStructFrom: fileStream)].
	"typecnt timesRepeat: [ttinfoStructs add: (self ttinfoStruct64From: fileStream)]."

	"Read time zone name abbreviations"
	fileStream atEnd ifTrue: [self error: 'unexpected end of file'].
	abbreviations := fileStream next: charcnt.
	"Note: tzNames is redundant, as the same information will be saved in the transitionTimeTable.
	The names are stored separately here as a convenience for browsing time zones."
	tzNames := self getTzNameAbbreviationsFrom: abbreviations.

	"Read leap second information"
	leapSeconds := OrderedCollection new.
	leapcnt timesRepeat: [leapSeconds add: (self leapSecondStruct64From: fileStream)].

	"Flags for standard/wall indicators, to indicate if transition times associated with
	local time types  were specified as standard  time or wall clock time, used
	when a time zone file is used in handling POSIX-style time zone environment
	variables."
	isStd := OrderedCollection new: typecnt.
	ttisstdcnt timesRepeat: [isStd add: (self nextBooleanCharFrom: fileStream)].
	"Store the flags as element 4 of the corresponding ttinfoStruct"
	(1 to: ttisstdcnt) do: [:e | (ttinfoStructs at: e) at: 4 put: (isStd at: e)].

	"Flags for UTC/local indicators, tell whether the transition times associated with
	local time types were specified as UTC or local time, used when a time zone file
	is used in handling POSIX-style time zone environment variables."
	isGmt := OrderedCollection new: typecnt.
	ttisgmtcnt timesRepeat: [isGmt add: (self nextBooleanCharFrom: fileStream)].
	"Store the flags as element 5 of the corresponding ttinfoStruct"
	(1 to: ttisstdcnt) do: [:e | (ttinfoStructs at: e) at: 5 put: (isGmt at: e)].

	"Read a newline-enclosed, POSIX-TZ-environment-variable-style string for use in
	handling instants after the last transition time stored in the file (an empty string
	if there is no POSIX representation for such instants)."
	posixTzEnvString := fileStream upToEnd.
	posixTzEnvString
		ifNotNil: [(posixTzEnvString first = Character lf asciiValue
					and: [posixTzEnvString last = Character lf asciiValue])
				ifFalse: [self notify: 'expected TZ environment string surrounded by Character lf'.
					^ nil].
			posixTzEnvString := posixTzEnvString asString allButFirst allButLast].

	"Relate the transition times to the ttinfoStruct data"
	transitionTable := Array new: timecnt.
	(1 to: timecnt) do: [:idx |
		rec := ttinfoStructs at: ((ttinfoIndices at: idx) + 1).
		transitionTable at: idx put:
			(TimeZoneRule new
				transitionTime: (transitionTimes at: idx);
				offsetSeconds: (rec at: 1);
				isDstFlag: (rec at: 2);
				abbreviation: (((ReadStream on: abbreviations)
								next: (rec at: 3); yourself) upTo: 0) asString asSymbol;
				isStdTimeTransition: (rec at: 4);
				isUtcTransition: (rec at: 5))].

	"Create an instance of LeapSecondRuleSet, or nil if no rules were found in this tzfile."
	(leapSeconds size = 0)
		ifTrue: [leapSecondRuleSet := nil]
		ifFalse: [leapSecondRuleSet := LeapSecondRuleSet new leapSecondTable: leapSeconds].

	"Create an instance of TimeZoneRuleSet"
	timeZoneRuleSet := TimeZoneRuleSet new.
	timeZoneRuleSet timeZoneName: name.
	timeZoneRuleSet transitionTimeTable: transitionTable.
	timeZoneRuleSet leapSecondRuleSet: leapSecondRuleSet.
	timeZoneRuleSet tzNameAbbreviations: tzNames asArray.
	timeZoneRuleSet posixTzEnvString: posixTzEnvString.
	^ Array with: timeZoneRuleSet
		with: leapSecondRuleSet
		with: versionOneData
! !

!TzFileLoader methodsFor: 'tzfile decoding' stamp: 'dtl 12/31/1999 18:21'!
magic: aStream
	"Answer the first four characters, which are the magic number for the tzfile file."

	aStream atEnd ifTrue: [self error: 'unexpected end of file'].
	^ (aStream next: 4) asString
! !

!TzFileLoader methodsFor: 'accessing' stamp: 'dtl 12/28/1999 20:01'!
name

	^ name! !

!TzFileLoader methodsFor: 'accessing' stamp: 'dtl 12/28/1999 20:01'!
name: aString

	name := aString! !

!TzFileLoader methodsFor: 'tzfile decoding' stamp: 'dtl 12/31/1999 17:12'!
nextBooleanCharFrom: aStream

	aStream atEnd ifTrue: [self error: 'unexpected end of file'].
	^ (aStream next asInteger) ~= 0
! !

!TzFileLoader methodsFor: 'tzfile decoding' stamp: 'dtl 5/31/2009 09:15'!
nextInt64FromNetworkOrderedBytes: aStream
	"Convert the next eight characters, read in network order, into an Integer.
	Logic is derived from ByteArray>>longAt:bigEndian:"

	| b0 b1 b2 b3 b4 b5 b6 w h |
	b0 := aStream next asInteger.
	b1 := aStream next asInteger.
	b2 := aStream next asInteger.
	b3 := aStream next asInteger.
	b4 := aStream next asInteger.
	b5 := aStream next asInteger.
	b6 := aStream next asInteger.
	w :=  aStream next asInteger.

	h := ((b0 bitAnd: 16r7F) - (b0 bitAnd: 16r80) bitShift: 8) + b1.
	b6 = 0 ifFalse:[w := (b6 bitShift: 8) + w].
	b5 = 0 ifFalse:[w := (b5 bitShift: 16) + w].
	b4 = 0 ifFalse:[w := (b4 bitShift: 24) + w].
	b3 = 0 ifFalse:[w := (b3 bitShift: 32) + w].
	b2 = 0 ifFalse:[w := (b2 bitShift: 40) + w].
	h = 0 ifFalse:[w := (h bitShift: 48) + w].
	^w
! !

!TzFileLoader methodsFor: 'tzfile decoding' stamp: 'dtl 12/31/1999 17:10'!
nextIntFromNetworkOrderedBytes: aStream
	"Convert the next four characters, read in network order, into an Integer.
	Logic is copied from ByteArray>>longAt:bigEndian:"

	| b0 b1 b2 w h |
	b0 := aStream next asInteger.
	b1 := aStream next asInteger.
	b2 := aStream next asInteger.
	w :=  aStream next asInteger.

	"Minimize LargeInteger arithmetic"
	h := ((b0 bitAnd: 16r7F) - (b0 bitAnd: 16r80) bitShift: 8) + b1.
	b2 = 0 ifFalse:[w := (b2 bitShift: 8) + w].
	h = 0 ifFalse:[w := (h bitShift: 16) + w].
	^w
! !

!TzFileLoader methodsFor: 'tzfile decoding' stamp: 'dtl 12/31/1999 17:12'!
nextUnsignedCharFrom: aStream

	aStream atEnd ifTrue: [self error: 'unexpected end of file'].
	^ aStream next asInteger
! !

!TzFileLoader methodsFor: 'tzfile decoding' stamp: 'dtl 12/31/1999 17:08'!
nextUnsignedIntFromNetworkOrderedBytes: aStream
	"Convert the next four characters, read in network order, into an Integer"

	aStream atEnd ifTrue: [self error: 'unexpected end of file'].
	^ (aStream next asInteger * 16777216) +
		(aStream next asInteger * 65536) +
		(aStream next asInteger * 256) +
		aStream next asInteger
! !

!TzFileLoader methodsFor: 'tzfile decoding' stamp: 'dtl 12/28/1999 20:01'!
ttinfoStructFrom: aStream

	| result |
	result := Array new: 5.
	result at: 1 put: (self nextIntFromNetworkOrderedBytes: aStream).
	result at: 2 put: (self nextUnsignedCharFrom: aStream).
	aStream atEnd ifTrue: [self error: 'unexpected end of file'].
	result at: 3 put: (self nextUnsignedCharFrom: aStream).
	^ result
! !
TimeZoneDatabase initialize!
